/**
 * Tribe: Group communication library.
 * Copyright (C) 2004-2005 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.adapters;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import org.objectweb.tribe.channel.ReliableGroupChannel;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.exceptions.TimeoutException;
import org.objectweb.tribe.messages.MessageListener;

/**
 * This class defines a MulticastRequestAdapter which is similar to JGroups
 * MessageDispatcher. It is a sort of multicast RPC (Remote Procedure Call)
 * where a message is multicasted and each member return a response.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class MulticastRequestAdapter implements MessageListener
{
  /** Do not wait for any response when multicasting a message (asynchronous) */
  public static final int          WAIT_NONE       = 0;
  /** Wait for the first reponse before returning of a multicast send */
  public static final int          WAIT_FIRST      = 1;
  /** Wait for a majority of responses before returning of a multicast send */
  public static final int          WAIT_MAJORITY   = 2;
  /** Wait for all responses before returning of a multicast send */
  public static final int          WAIT_ALL        = 3;

  private ReliableGroupChannel     channel;
  private Member                   me;
  private MessageListener          msgListener;
  private MulticastRequestListener multicastRequestListener;
  private HashMap                  pendingQueries;
  private PullPushAdapter          pullPushAdapter = null;
  private int                      sequenceNumber  = 0;
  private static Trace             logger          = Trace
                                                       .getLogger("org.objectweb.tribe.blocks.multicastadapter");

  /**
   * Creates a new <code>MulticastRequestAdapter</code> object
   */
  public MulticastRequestAdapter(ReliableGroupChannel channel,
      MessageListener msgListener, MulticastRequestListener dispatcherListener)
  {
    pullPushAdapter = new PullPushAdapter(channel, this);
    this.channel = channel;
    this.msgListener = msgListener;
    this.multicastRequestListener = dispatcherListener;
    this.me = channel.getLocalMembership();
    this.pendingQueries = new HashMap();
  }

  /**
   * Returns the channel value.
   * 
   * @return Returns the channel.
   */
  public ReliableGroupChannel getChannel()
  {
    return channel;
  }

  /**
   * Stop this MulticastRequestAdapter. This does not close the underlying
   * channel but terminates the PullPushAdapter create for this
   * MulticastRequestAdapter.
   */
  public void stop()
  {
    pullPushAdapter.stop();
  }

  /**
   * @see org.objectweb.tribe.messages.MessageListener#receive(java.io.Serializable)
   */
  public void receive(Serializable msg)
  {
    if (msg instanceof MulticastRequestAdapterMessage)
    {
      MulticastRequestAdapterMessage requestReply = (MulticastRequestAdapterMessage) msg;
      if (requestReply.isReply())
      { // It's a reply
        MulticastResponse resp;
        Integer key = new Integer(requestReply.getUid());
        synchronized (pendingQueries)
        {
          resp = (MulticastResponse) pendingQueries.get(key);
        }
        if (resp == null)
        { // No MulticastReponse object, it has probably been removed because
          // wait mode policy is not synchronous.
          if (logger.isDebugEnabled())
            logger.debug("Dropping response to request "
                + requestReply.getUid());
          return;
        }
        // Add the reply
        if (logger.isDebugEnabled())
          logger.debug("Received reply from " + requestReply.getSender()
              + " to message " + requestReply.getUid());
        resp.addResult(requestReply.getSender(), requestReply.getMessage());
      }
      else
      { // It's a request, forward first to single threaded callback
        Object callback1 = multicastRequestListener
            .handleMessageSingleThreaded(requestReply.getMessage(),
                requestReply.getSender());

        // Now, let the message be handled in a separate thread
        new MulticastRequestAdapterThread(multicastRequestListener,
            requestReply, callback1, pullPushAdapter, me).start();
      }
    }
    else
      // Regular message, forward to listener
      msgListener.receive(msg);
  }

  /**
   * Multicast a message to the given list of member.
   * 
   * @param dests members to send the message to1
   * @param msg message to send
   * @param waitMode one of WAIT_NONE, WAIT_FIRST, WAIT_MAJORITY or WAIT_ALL
   * @param timeout time in ms to wait for responses (0 means no timeout)
   * @return the response list
   * @throws TimeoutException if the timeout expires
   * @throws NotConnectedException if the underlying channel is not connected to
   *           a group
   * @throws ChannelException if an error occurs during transmission
   */
  public MulticastResponse multicastMessage(ArrayList dests, Serializable msg,
      int waitMode, long timeout) throws TimeoutException, ChannelException,
      NotConnectedException
  {
    // Create request
    int seq = nextSequenceNumber();
    // Object for collecting responses
    MulticastResponse reply = new MulticastResponse(dests, waitMode);
    MulticastRequestAdapterMessage toSend;
    if (waitMode == WAIT_NONE)
    { // No answer required, just report failed nodes
      toSend = new MulticastRequestAdapterMessage(msg, me, seq,
          MulticastRequestAdapterMessage.REQUEST_ONLY);
      if (logger.isDebugEnabled())
        logger.debug("Sending request " + seq + " to " + dests.size()
            + " members.");
      ArrayList failed = pullPushAdapter.send(toSend, dests);
      reply.setFailedMembers(failed);
      return reply;
    }
    else
    { // We are waiting for an answer
      toSend = new MulticastRequestAdapterMessage(msg, me, seq,
          MulticastRequestAdapterMessage.REQUEST);

      Integer key = new Integer(seq);
      synchronized (pendingQueries)
      {
        pendingQueries.put(key, reply);
      }

      // Send the message and wait for responses
      if (logger.isDebugEnabled())
        logger.debug("Sending message " + key + " to " + dests.size()
            + " members.");
      ArrayList failed = pullPushAdapter.send(toSend, dests);
      reply.setFailedMembers(failed);
      reply.waitForCompletion(timeout);

      // Release the MulticastResponse object
      synchronized (pendingQueries)
      {
        pendingQueries.remove(key);
      }
      return reply;
    }
  }

  private synchronized int nextSequenceNumber()
  {
    return sequenceNumber++;
  }

}