/**
 * Tribe: Group communication library.
 * Copyright (C) 2002-2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.adapters;

import java.io.Serializable;

import org.objectweb.tribe.common.Member;

/**
 * This class defines a MulticastRequestAdapterMessage for carrying request and
 * replies of multicastMessage().
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
class MulticastRequestAdapterMessage implements Serializable
{
  static final int     REQUEST      = 0;
  static final int     REQUEST_ONLY = 1;
  static final int     REPLY        = 2;
  private Serializable msg;
  private Member       sender;
  private int          uid;
  private int          requestReply;

  /**
   * Creates a new <code>MulticastRequestAdapterMessage</code> object
   * 
   * @param msg message to send
   * @param sender message sender
   * @param message unique identifier
   * @param requestOrReply REQUEST, REQUEST_ONLY or REPLY
   */
  public MulticastRequestAdapterMessage(Serializable msg, Member sender,
      int uid, int requestOrReply)
  {
    this.msg = msg;
    this.sender = sender;
    this.uid = uid;
    this.requestReply = requestOrReply;
  }

  /**
   * Returns the message value.
   * 
   * @return Returns the message.
   */
  public Serializable getMessage()
  {
    return msg;
  }

  /**
   * Sets the message value.
   * 
   * @param msg The message to set.
   */
  public void setMessage(Serializable msg)
  {
    this.msg = msg;
  }

  /**
   * Returns true if the message is a request.
   * 
   * @return true if the message is a request.
   */
  public boolean isRequest()
  {
    return requestReply == REQUEST;
  }

  /**
   * Sets this message to a request.
   */
  public void setRequest()
  {
    this.requestReply = REQUEST;
  }

  /**
   * Returns true if the message is a request only.
   * 
   * @return true if the message is a request only.
   */
  public boolean isRequestOnly()
  {
    return requestReply == REQUEST_ONLY;
  }

  /**
   * Sets this message to a request only.
   */
  public void setRequestOnly()
  {
    this.requestReply = REQUEST_ONLY;
  }

  /**
   * Returns true if the message is a reply.
   * 
   * @return true if the message is a reply.
   */
  public boolean isReply()
  {
    return requestReply == REPLY;
  }

  /**
   * Sets this message to a reply.
   */
  public void setReply()
  {
    this.requestReply = REPLY;
  }

  /**
   * Returns the sender of the message.
   * 
   * @return Returns the sender.
   */
  public Member getSender()
  {
    return sender;
  }

  /**
   * Sets the sender of the message.
   * 
   * @param sender The sender to set.
   */
  public void setSender(Member sender)
  {
    this.sender = sender;
  }

  /**
   * Returns the uid value.
   * 
   * @return Returns the uid.
   */
  public int getUid()
  {
    return uid;
  }

  /**
   * @see java.lang.Object#toString()
   */
  public String toString()
  {
    String type;
    switch (requestReply)
    {
      case REQUEST :
        type = "Request ";
        break;
      case REQUEST_ONLY :
        type = "Request only ";
        break;
      case REPLY :
        type = "Reply ";
        break;
      default :
        type = "Unknown ";
        break;
    }
    return type + uid + " from " + sender + " (" + msg + ")";
  }
}