/**
 * Tribe: Group communication library.
 * Copyright (C) 2002-2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.adapters;

import java.io.Serializable;
import java.util.ArrayList;

import org.objectweb.tribe.common.Member;

/**
 * This class defines a MulticastRequestListener to handle calls from a
 * MulticastRequestAdapter. When MulticastRequestAdapter.multicastMessage(...)
 * is called, the remote node must reply (possibly with a null message). The
 * message handling is split in 2 callbacks:
 * <p>
 * 1. call to handleMessageSingleThreaded(...). This is done in mutual exclusion
 * and no other message can be delivered until this function has returned. An
 * object can be returned to pass temporary data to the second handler.
 * <p>
 * 2. handleMessageMultiThreaded is executed in a dedicated thread and return
 * the result of the call in a Serializable object.
 * <p>
 * Note that any of these handlers might be empty and just return null.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public interface MulticastRequestListener
{

  /**
   * First callback to handle a message sent through
   * MulticastRequestAdapter.multicastMessage(...). This method is executed in
   * mutual exclusion and block subsequent message deliveries until this
   * callback returns. The object returned by this function is used to pass
   * temporary data to the second callback.
   * 
   * @param msg the message received that needs to be handled
   * @param sender the message sender
   * @return the object to pass to handleMessageMultiThreaded
   * @see MulticastRequestAdapter#multicastMessage(ArrayList, Serializable, int,
   *      long)
   * @see #handleMessageMultiThreaded(Serializable, Member, Object)
   */
  Object handleMessageSingleThreaded(Serializable msg, Member sender);

  /**
   * Second callback to handle a message sent through
   * MulticastRequestAdapter.multicastMessage(...). The code of this callback
   * will execute in a dedicated thread and therefore will execute concurrently
   * of other messages processing.
   * 
   * @param msg the message received that needs to be handled
   * @param sender the message sender
   * @param handleMessageSingleThreadedResult result of the first callback (may
   *          be null)
   * @return the reply to send back
   * @see MulticastRequestAdapter#multicastMessage(ArrayList, Serializable, int,
   *      long)
   */
  Serializable handleMessageMultiThreaded(Serializable msg, Member sender,
      Object handleMessageSingleThreadedResult);
}