/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.adapters;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.exceptions.TimeoutException;

/**
 * This class defines a MulticastResponse
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class MulticastResponse
{
  private ArrayList recipient;
  private ArrayList succeedMembers;
  private ArrayList failedMembers;
  private HashMap   results;
  private int       waitMode;
  private int       nbOfRecipients;

  /**
   * Creates a new <code>MulticastResponse</code> object
   * 
   * @param recipient <code>ArrayList</code> of <code>Member</code> to which
   *          the request is sent
   * @param waitMode one of MulticastRequestAdapter.WAIT_NONE, WAIT_FIRST,
   *          WAIT_MAJORITY or WAIT_ALL
   */
  public MulticastResponse(ArrayList recipient, int waitMode)
  {
    this.recipient = recipient;
    this.waitMode = waitMode;
    nbOfRecipients = recipient.size();
    succeedMembers = new ArrayList(nbOfRecipients);
    results = new HashMap(nbOfRecipients);
  }

  /**
   * Returns the list of failed Members.
   * 
   * @return Returns the failed Members.
   */
  public ArrayList getFailedMembers()
  {
    return failedMembers;
  }

  /**
   * Sets the list of failed Members. If the list of failed members is not null
   * then we check if the wait mode policy has been satisfied and unblock
   * waitForCompletion if needed.
   * 
   * @param failedMembers The failed Members to set.
   */
  public void setFailedMembers(ArrayList failedMembers)
  {
    synchronized (results)
    {
      this.failedMembers = failedMembers;
      if (failedMembers != null)
        if (shouldNotify())
          results.notify();
    }
  }

  /**
   * Returns the member results (for all Members that succeeded). The Hashmap
   * uses the Member as a key and the result as a value.
   * 
   * @return Returns the member results.
   */
  public HashMap getResults()
  {
    return results;
  }

  /**
   * Get the result returned by the specified member.
   * 
   * @param m the member for which we need the result
   * @return the result or null if no result has been found (member failure or
   *         member not part of this request).
   */
  public Serializable getResult(Member m)
  {
    return (Serializable) results.get(m);
  }

  /**
   * Add the given result for the specified member.
   * <p>
   * The member is then added in the succeedMembers list and the
   * waitForCompletion() method might be unblocked according to the waiting mode
   * policy specified in the constructor.
   * 
   * @param m member sending this result
   * @param result result to add.
   */
  public void addResult(Member m, Serializable result)
  {
    synchronized (results)
    {
      results.put(m, result);
      succeedMembers.add(m);
      if (shouldNotify())
        results.notify();
    }
  }

  /**
   * Returns the list of members who sucessfully returned a response.
   * 
   * @return Returns the succeed Members.
   * @see #addResult(Member, Serializable)
   */
  public ArrayList getSucceedMembers()
  {
    return succeedMembers;
  }

  /**
   * Returns the original list of the recipients (the ones we sent the message
   * to).
   * 
   * @return Returns an <code>ArrayList</code> of <code>Member</code>.
   */
  public ArrayList getRecipient()
  {
    return recipient;
  }

  /**
   * Wait until the needed number of responses has been received (according to
   * the waiting mode policy).
   * 
   * @param timeout maximum time in ms to wait for (0 means no timeout)
   */
  public void waitForCompletion(long timeout) throws TimeoutException
  {
    long start = 0;
    long end = 0;
    synchronized (results)
    {
      if (shouldNotify())
        return;
      try
      {
        start = System.currentTimeMillis();
        results.wait(timeout);
        end = System.currentTimeMillis();
      }
      catch (InterruptedException ignore)
      {
      }
    }
    if ((timeout != 0) && (end - start >= timeout))
      throw new TimeoutException();
  }

  /**
   * Check if waitForCompletion should be notified of completion. This method
   * should be calls while synchronized(results).
   * 
   * @return true if the wait mode policy has been satisfied and the expected
   *         responses have been received.
   */
  private boolean shouldNotify()
  {
    int succeedSize = succeedMembers.size();
    switch (waitMode)
    {
      case MulticastRequestAdapter.WAIT_NONE :
        return true;
      case MulticastRequestAdapter.WAIT_FIRST :
        return (succeedSize > 0)
            || ((failedMembers != null) && (failedMembers.size() == nbOfRecipients));
      case MulticastRequestAdapter.WAIT_MAJORITY :
        return (succeedSize > (nbOfRecipients / 2))
            || ((failedMembers != null) && (succeedSize + failedMembers.size() == nbOfRecipients));
      case MulticastRequestAdapter.WAIT_ALL :
        return (succeedSize == nbOfRecipients)
            || ((failedMembers != null) && (succeedSize + failedMembers.size() == nbOfRecipients));
      default :
        throw new RuntimeException(
            "Unexpected wait mode policy in MulticastReponse: " + waitMode);
    }
  }

}