/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): __________________.
 */

package org.objectweb.tribe.adapters;

import java.io.Serializable;
import java.util.ArrayList;

import org.objectweb.tribe.channel.ReliableGroupChannel;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.messages.MessageListener;

/**
 * Allows a client to be notified when messages have been received instead of
 * having to actively poll the channel for new messages. Typically used in the
 * client role (receive()). This blocks works with any kind of reliable group
 * channel.
 * <p>
 * Note that message delivery is single threaded and if the message handler
 * hangs, it will block further message delivery. Send() methods are just
 * provided for convenience and wraps calls to the same methods on the channel.
 * <p>
 * This adapter is inspired from JGroups PullPushAdapter but does not deal with
 * views.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @author Bela Ban
 * @version 1.0
 */
public class PullPushAdapter implements Runnable
{
  protected ReliableGroupChannel channel      = null;
  protected MessageListener      listener     = null;
  protected Thread               readerThread = null;
  private static Trace           logger       = Trace
                                                  .getLogger("org.objectweb.tribe.channel");
  private boolean                isKilled;

  /**
   * Creates a new <code>PullPushAdapter</code> on a channel and use the
   * provided listener for callbacks on message reception.
   * 
   * @param channel channel to read messages from
   * @param listener MessageListener for callbacks
   */
  public PullPushAdapter(ReliableGroupChannel channel, MessageListener listener)
  {
    this.channel = channel;
    this.listener = listener;
    start();
  }

  /**
   * Return the underlying channel.
   * 
   * @return underlying channel
   */
  public ReliableGroupChannel getChannel()
  {
    return channel;
  }

  /**
   * Start a new thread
   */
  public void start()
  {
    if (readerThread == null)
    {
      readerThread = new Thread(this, "PullPushAdapterThread");
      isKilled = false;
      readerThread.start();
      Thread.yield(); // Let the thread start
    }
  }

  /**
   * Stop the current adapter and its associated thread.
   */
  public void stop()
  {
    if ((readerThread != null) && readerThread.isAlive())
    {
      isKilled = true;
      readerThread.interrupt();
      try
      {
        channel.send(null);
      }
      catch (Exception ignore)
      {
      }
    }
    try
    {
      readerThread.join(1000);
    }
    catch (Exception ex)
    {
    }
  }

  /**
   * Sends the given message to all members of the group.
   * 
   * @param msg message to send
   * @return an <code>ArrayList</code> of Members who failed, or null if all
   *         members received successfully the message.
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   * @see ReliableGroupChannel#send(Serializable)
   */
  public ArrayList send(Serializable msg) throws ChannelException,
      NotConnectedException
  {
    return channel.send(msg);
  }

  /**
   * Sends a message to a subset of group members.
   * 
   * @param msg message to send
   * @param members <code>ArrayList</code> of <code>Member</code> that are
   *          all part of the group
   * @return an <code>ArrayList</code> of Members who failed, or null if all
   *         members received successfully the message.
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   * @see ReliableGroupChannel#send(Serializable, ArrayList)
   */
  public ArrayList send(Serializable msg, ArrayList members)
      throws ChannelException, NotConnectedException
  {
    return channel.send(msg, members);
  }

  /**
   * Sends a message to members of group gid without being necessary member of
   * the group.
   * 
   * @param msg message to send
   * @param members <code>ArrayList</code> of <code>Member</code> that are
   *          all part of the group
   * @return an <code>ArrayList</code> of Members who failed, or null if all
   *         members received successfully the message.
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   * @see ReliableGroupChannel#send(Serializable, GroupIdentifier, ArrayList)
   */
  public ArrayList send(Serializable msg, GroupIdentifier gid, ArrayList members)
      throws ChannelException, NotConnectedException
  {
    return channel.send(msg, gid, members);
  }

  /**
   * Reentrant run(): message reception is serialized, then the listener is
   * notified of the message reception
   */
  public void run()
  {
    Serializable msg;

    while (!isKilled)
    {
      try
      {
        msg = channel.receive();
        listener.receive(msg);
      }
      catch (Exception e)
      {
        logger.debug("PullPushAdapter: Error while reading from channel", e);
      }
    }
  }

}