/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;

import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.messages.ChannelMessage;

/**
 * This class defines a AbstractChannelPool
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public abstract class AbstractChannelPool
{
  // HashMap: IpAddress -> AbstractReliableFifoChannel
  protected HashMap    channels;
  // HashMap: IpAddress -> AbstractServerChannel
  protected HashMap    serverChannels;
  // HashMap: Object(keys) -> ArrayList<ReceiveBuffer>
  protected HashMap    keyBuffers;

  private static Trace logger = Trace.getLogger("org.objectweb.tribe.channel");

  /**
   * Creates a new <code>AbstractChannelPool</code> object
   */
  public AbstractChannelPool()
  {
    channels = new HashMap();
    serverChannels = new HashMap();
    keyBuffers = new HashMap();
  }

  /**
   * Get a reliable fifo serverSocket to the given destination.
   * <p>
   * If no serverSocket exist, a new one is created else an existing one may be
   * returned.
   * 
   * @param destination destination address to reach
   * @return a reliable fifo serverSocket
   * @throws ChannelException if an error occurs
   */
  public abstract AbstractReliableFifoChannel getChannel(Address destination)
      throws ChannelException;

  /**
   * Removed the specified channel from the pool.
   * 
   * @param channel the channel to remove
   * @return true if the channel was removed from the pool, false if the channel
   *         wasn't found in the pool.
   */
  public boolean removeChannelFromPool(AbstractReliableFifoChannel channel)
  {
    synchronized (channels)
    {
      Collection values = channels.values();
      if (values == null)
        return false;
      return values.remove(channel);
    }
  }

  /**
   * Get a server socket on the given address.
   * <p>
   * If no server Socket exist, a new one is created else an existing one may be
   * returned.
   * 
   * @param serverAddress server address to bind
   * @return a server serverSocket
   * @throws ChannelException if an error occurs
   */
  public abstract AbstractServerChannel getServerChannel(Address serverAddress)
      throws ChannelException;

  /**
   * Removed the specified channel from the pool.
   * 
   * @param channel the channel to remove
   * @return true if the channel was removed from the pool, false if the channel
   *         wasn't found in the pool.
   */
  public boolean removeServerChannelFromPool(AbstractServerChannel channel)
  {
    synchronized (serverChannels)
    {
      Collection values = serverChannels.values();
      if (values == null)
        return false;
      return values.remove(channel);
    }
  }

  /**
   * Register a new receive buffer (which includes the group identifier the
   * client is interested in).
   * 
   * @param buffer ReceiveBuffer to register
   */
  public void registerReceiveBuffer(ReceiveBuffer buffer)
  {
    Object key = buffer.getBufferKey();
    synchronized (keyBuffers)
    {
      if (logger.isDebugEnabled())
        logger.debug("Registering new receive buffer: " + key);
      // Get buffers for this gid
      ArrayList buffers = (ArrayList) keyBuffers.get(key);
      if (buffers == null)
      { // None, create a new list
        buffers = new ArrayList();
        keyBuffers.put(key, buffers);
      }
      // Add the given buffer
      synchronized (buffers)
      {
        buffers.add(buffer);
      }
    }
  }

  /**
   * Unregister a receive buffer. When the last receive buffer is unregistered,
   * all channels are closed.
   * 
   * @param buffer ReceiveBuffer to unregister
   * @return true if the ReceiveBuffer was successfully unregistered, false if
   *         it was not found in this pool.
   */
  public boolean unregisterReceiveBuffer(ReceiveBuffer buffer)
  {
    Object key = buffer.getBufferKey();
    synchronized (keyBuffers)
    {
      if (logger.isDebugEnabled())
        logger.debug("Unregistering receive buffer: " + key);
      // Get buffers for this gid
      ArrayList buffers = (ArrayList) keyBuffers.get(key);
      if (buffers == null)
        return false; // No buffer for this group id
      // Remove the given buffer
      synchronized (buffers)
      {
        boolean wasUnregistered = buffers.remove(buffer);
        if (buffer.isEmpty())
        { // Close all channels, no subscription to anything anymore
          Object[] channelsToClose = channels.values().toArray();
          for (int i = 0; i < channelsToClose.length; i++)
          {
            AbstractReliableFifoChannel channel = (AbstractReliableFifoChannel) channelsToClose[i];
            removeChannelFromPool(channel);
          }
          channels.clear();
        }
        if (wasUnregistered)
          buffer.addMessage(null); // Notify the queue to wake up blocking
        // threads
        return wasUnregistered;
      }
    }
  }

  /**
   * Sends a message to the given list of recipients.
   * 
   * @param msg message to send
   * @param members list of destination members
   * @return list of members who failed, null if everything was successful.
   */
  public ArrayList send(ChannelMessage msg, ArrayList members)
  {
    ArrayList failed = null;
    Address lastAddress = null;
    synchronized (members)
    {
      int size = members.size();
      for (int i = 0; i < size; i++)
      {
        Member m = (Member) members.get(i);
        // Check that we do not send multiple times to co-located members
        if (!m.getAddress().equals(lastAddress))
        {
          AbstractReliableFifoChannel channel = null;
          try
          {
            channel = getChannel(m.getAddress());
            channel.send(msg);
            lastAddress = m.getAddress();
          }
          catch (Exception e)
          {
            logger.debug("Failed to send message to member " + m, e);
            if (failed == null)
              failed = new ArrayList();
            failed.add(m);
            removeChannelFromPool(channel);
            lastAddress = null;
          }
        }
      }
    }
    return failed;
  }

}