/**
 * C-JDBC: Clustered JDBC.
 * Copyright (C) 2002-2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: c-jdbc@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel;

import java.io.Serializable;
import java.net.URL;
import java.util.ArrayList;

import org.jgroups.Address;
import org.jgroups.ChannelClosedException;
import org.jgroups.ChannelNotConnectedException;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.TimeoutException;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.messages.GroupMessage;

/**
 * This class defines a JGroupsReliableChannel which is a ReliableGroupChannel
 * wrapper on top of a JGroups JChannel.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class JGroupsReliableChannel extends ReliableGroupChannel
{
  // The underlying JGroups channel
  private JChannel jgroupsChannel;

  /**
   * Creates a new <code>JGroupsReliableChannel</code> object
   */
  public JGroupsReliableChannel(URL jgroupsConfigFile) throws ChannelException
  {
    try
    {
      jgroupsChannel = new JChannel(jgroupsConfigFile);
    }
    catch (org.jgroups.ChannelException e)
    {
      throw new ChannelException(e);
    }
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#join(org.objectweb.tribe.common.Group)
   */
  public void join(Group g) throws AlreadyMemberException, ChannelException,
      NotConnectedException
  {
    if ((currentGroup != null) && currentGroup.equals(g))
      throw new AlreadyMemberException();
    // Quit the previous group first if needed
    if (currentGroup != null)
      try
      {
        quit();
      }
      catch (Exception ignore)
      {
      }
    currentGroup = g;
    try
    {
      jgroupsChannel.connect(g.getGroupIdentifier().getGroupName());
    }
    catch (ChannelClosedException e)
    {
      throw new NotConnectedException(e);
    }
    catch (org.jgroups.ChannelException e)
    {
      throw new ChannelException(e);
    }
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#quit()
   */
  public void quit() throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    jgroupsChannel.close();
    currentGroup = null;
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#receive()
   */
  public Serializable receive() throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    try
    {
      GroupMessage groupMessage = (GroupMessage) jgroupsChannel.receive(0);
      if (groupMessage == null)
        return null;
      else
        return groupMessage.getMessage();
    }
    catch (ChannelNotConnectedException e)
    {
      throw new NotConnectedException(e);
    }
    catch (ChannelClosedException e)
    {
      throw new NotConnectedException(e);
    }
    catch (TimeoutException e)
    {
      throw new ChannelException(
          "Timeout while retrieving message from channel", e);
    }
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#send(java.io.Serializable,
   *      org.objectweb.tribe.common.GroupIdentifier, java.util.ArrayList)
   */
  public ArrayList send(Serializable msg, GroupIdentifier gid, ArrayList members)
      throws ChannelException, NotConnectedException
  {
    Address dest = null;
    Message jgroupsMsg = new Message(dest, null, new GroupMessage(msg, gid));
    try
    {
      jgroupsChannel.send(jgroupsMsg);
      // Note that JGroups channel does not offer the facility of finding which
      // node have failed, it will be reported by a view change
      return null;
    }
    catch (ChannelNotConnectedException e)
    {
      throw new NotConnectedException(e);
    }
    catch (ChannelClosedException e)
    {
      throw new NotConnectedException(e);
    }
  }

  /**
   * @see java.lang.Object#toString()
   */
  public String toString()
  {
    return "JGroups channel wrapper: " + jgroupsChannel;
  }

  /**
   * Return the JGroups channel properties
   * 
   * @return JGroups channel properties
   */
  public String getProperties()
  {
    return jgroupsChannel.getProperties();
  }
}