/**
 * C-JDBC: Clustered JDBC.
 * Copyright (C) 2002-2005 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: c-jdbc@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel;

import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import org.jgroups.Address;
import org.jgroups.BlockEvent;
import org.jgroups.ChannelClosedException;
import org.jgroups.ChannelNotConnectedException;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.SuspectEvent;
import org.jgroups.TimeoutException;
import org.jgroups.View;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.gms.JGroupsMembershipService;
import org.objectweb.tribe.messages.FragmentedMessage;
import org.objectweb.tribe.messages.GroupMessage;

/**
 * This class defines a JGroupsReliableChannelWithGms which is a
 * ReliableGroupChannelWithGms wrapper on top of a JGroups JChannel that uses
 * the GMS service provided by JGroups.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class JGroupsReliableChannelWithGms extends ReliableGroupChannelWithGms
{
  // The underlying JGroups channel
  private JChannel                 jgroupsChannel;
  private JGroupsMembershipService jgroupsGMS;

  // Fragment size for oversized messages
  private static final int         FRAGMENT_SIZE = 32000;
  private HashMap                  fragmentList  = new HashMap();

  /**
   * Creates a new <code>JGroupsReliableChannel</code> object
   * 
   * @param jGroupsChannel JGroups channel to use
   * @param gms the GMS service to use
   */
  public JGroupsReliableChannelWithGms(JGroupsMembershipService gms)
  {
    super(gms);
    this.jgroupsGMS = gms;
    this.jgroupsChannel = gms.getJGroupsChannel();
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#join(org.objectweb.tribe.common.Group)
   */
  public void join(Group g) throws AlreadyMemberException, ChannelException,
      NotConnectedException
  {
    if ((currentGroup != null) && currentGroup.equals(g))
      throw new AlreadyMemberException();

    // Quit the previous group first if needed
    if (currentGroup != null)
      try
      {
        quit();
      }
      catch (Exception ignore)
      {
      }

    currentGroup = g;
    me = jgroupsGMS.join(this, g.getGroupIdentifier());
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#quit()
   */
  public void quit() throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    jgroupsChannel.close();
    currentGroup = null;
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#receive()
   */
  public Serializable receive() throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    while (true)
    { // Loop to process/filter JGroups internal events
      try
      {
        Object obj = jgroupsChannel.receive(0);

        // JGroups "feature" that might deliver null messages for no reason (see
        // JGroups' PushPullAdapter code).
        if (obj == null)
          continue;

        if (obj instanceof View)
        {
          jgroupsGMS.viewAccepted((View) obj);
          continue;
        }
        else if (obj instanceof SuspectEvent)
        {
          jgroupsGMS.suspect((Address) ((SuspectEvent) obj).getMember());
          continue;
        }
        else if (obj instanceof BlockEvent)
        {
          jgroupsGMS.block();
          continue;
        }
        else if (obj instanceof Message)
        {
          Object content = ((Message) obj).getObject();
          GroupMessage groupMessage = null;
          if (content instanceof FragmentedMessage)
          {
            FragmentedMessage fragment = (FragmentedMessage) content;
            byte[] message;
            if (fragment.getFragmentId() == 0)
            { // First fragment, allocate message
              message = new byte[fragment.getMessageSize()];
              fragmentList.put(fragment.getMessageId(), message);
            }
            else
            { // Retrieve message
              message = (byte[]) fragmentList.get(fragment.getMessageId());
            }

            // Copy fragment
            int offset = fragment.getFragmentId() * FRAGMENT_SIZE;
            int fragmentLength = fragment.getFragmentData().length;
            System.arraycopy(fragment.getFragmentData(), 0, message, offset,
                fragmentLength);

            // Was it the last fragment?
            if (offset + fragmentLength == fragment.getMessageSize())
            { // Last fragment convert to a GroupMessage Object
              try
              {
                ObjectInputStream ois = new ObjectInputStream(
                    new ByteArrayInputStream(message));
                groupMessage = (GroupMessage) ois.readObject();
              }
              catch (Exception e)
              {
                e.printStackTrace();
                throw new ChannelException(
                    "Failed to reassemble fragmented message", e);
              }
              finally
              { // Remove message from list
                fragmentList.remove(fragment.getMessageId());
              }
            }
            else
            { // Not last fragment
              continue;
            }
          }
          else if (content instanceof GroupMessage)
            groupMessage = (GroupMessage) content;

          // Do not else if here, if the last fragment was read, obj has been
          // modified just above to contain the complete group message.
          if (groupMessage == null)
            return null;
          else
          { // Check that we should receive this message
            ArrayList members = (ArrayList) groupMessage.getChunks().get(1);
            if (members.contains(getLocalMembership()))
              return groupMessage.getMessage();
          }
        }
        else
          System.out.println("Unhandled JGroups message type ("
              + obj.getClass() + "): " + obj);
      }
      catch (ChannelNotConnectedException e)
      {
        throw new NotConnectedException(e);
      }
      catch (ChannelClosedException e)
      {
        throw new NotConnectedException(e);
      }
      catch (TimeoutException e)
      {
        throw new ChannelException(
            "Timeout while retrieving message from channel", e);
      }
    }
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#send(java.io.Serializable,
   *      org.objectweb.tribe.common.GroupIdentifier, java.util.ArrayList)
   */
  public ArrayList send(Serializable msg, GroupIdentifier gid, ArrayList members)
      throws ChannelException, NotConnectedException
  {
    // JGroups channel can only send to the whole group else unicast will bypass
    // the ordering layer. Therefore we embed the destination members in the
    // message for filtering on the receiver side.
    GroupMessage tribeMessage = new GroupMessage(msg, gid);
    tribeMessage.addChunk(members);
    int msgSize = tribeMessage.getByteArray().length;
    try
    {
      if (msgSize > FRAGMENT_SIZE)
      { // Fragmentation needed since JGroups does not do it properly
        byte[] completeMessageInBytes = tribeMessage.getByteArray();
        byte[] fragmentData = new byte[FRAGMENT_SIZE];
        FragmentedMessage fragment = new FragmentedMessage(msgSize);
        int nbOfFragments = msgSize / FRAGMENT_SIZE;
        int currentFragment = 0;
        while (currentFragment < nbOfFragments)
        {
          System.arraycopy(completeMessageInBytes, currentFragment
              * FRAGMENT_SIZE, fragmentData, 0, FRAGMENT_SIZE);
          fragment.setFragment(currentFragment, fragmentData);
          jgroupsChannel.send(new Message(null, null, fragment));
          currentFragment++;
        }
        // Send last fragment
        int lastFragmentSize = msgSize % FRAGMENT_SIZE;
        if (lastFragmentSize > 0)
        {
          // Adjust to the right size
          fragmentData = new byte[lastFragmentSize];
          System.arraycopy(completeMessageInBytes, currentFragment
              * FRAGMENT_SIZE, fragmentData, 0, lastFragmentSize);
          fragment.setFragment(currentFragment, fragmentData);
          jgroupsChannel.send(new Message(null, null, fragment));
        }
      }
      else
      { // No fragmentation
        Message jgroupsMsg = new Message(null, null, tribeMessage);
        jgroupsChannel.send(jgroupsMsg);
      }
      // Note that JGroups channel does not offer the facility of finding which
      // node have failed, it will be reported by a view change
      return null;
    }
    catch (ChannelNotConnectedException e)
    {
      throw new NotConnectedException(e);
    }
    catch (ChannelClosedException e)
    {
      throw new NotConnectedException(e);
    }
  }

  /**
   * @see java.lang.Object#toString()
   */
  public String toString()
  {
    return "JGroups channel wrapper: " + jgroupsChannel;
  }

  /**
   * Return the JGroups channel properties
   * 
   * @return JGroups channel properties
   */
  public String getProperties()
  {
    return jgroupsChannel.getProperties();
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#getLocalMembership()
   */
  public Member getLocalMembership()
  {
    return JGroupsMembershipService
        .memberFromJGroupsAddress((org.jgroups.stack.IpAddress) jgroupsChannel
            .getLocalAddress());
  }

  /**
   * @see org.objectweb.tribe.channel.ReliableGroupChannel#getCurrentGroup()
   */
  public Group getCurrentGroup()
  {
    if (currentGroup == null)
      return null;
    else
      return gms.getGroup(currentGroup.getGroupIdentifier());
  }

}