/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel;

import java.util.ArrayList;

import org.objectweb.tribe.exceptions.EmptyBufferException;
import org.objectweb.tribe.messages.ChannelMessage;

/**
 * This class defines a ReceiveBuffer which behaves as a FIFO buffer.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class ReceiveBuffer
{
  /** The queue holding messages */
  private ArrayList queue;
  private Object    bufferKey;

  /**
   * Creates a new <code>ReceiveBuffer</code> that collects messages sent to
   * the specified destination key.
   * 
   * @param bufferKey message destination key.
   */
  public ReceiveBuffer(Object bufferKey)
  {
    queue = new ArrayList();
    this.bufferKey = bufferKey;
  }

  /**
   * Returns the destination key this buffer is interested in.
   * 
   * @return Returns the buffer destination key.
   */
  public Object getBufferKey()
  {
    return bufferKey;
  }

  /**
   * Adds a new message to the buffer.
   * 
   * @param message Message to add
   */
  public void addMessage(Object message)
  {
    synchronized (queue)
    {
      // Note that it is safe to notify first since the awakened thread cannot
      // access the queue before we release the lock by exiting this method.
      if (queue.isEmpty())
        queue.notifyAll();
      queue.add(message);
    }
  }

  /**
   * Get the first message from the buffer. This method blocks until a message
   * is available from the buffer.
   * 
   * @return the first ChannelMessage
   * @throws EmptyBufferException if an error occurs
   */
  public ChannelMessage getMessage() throws EmptyBufferException
  {
    synchronized (queue)
    {
      while (queue.isEmpty())
      {
        try
        {
          queue.wait();
        }
        catch (InterruptedException ignore)
        {
        }
      }
      return (ChannelMessage) removeMessage();
    }
  }

  /**
   * Removes the first message from the buffer.
   * 
   * @return the first message from the queue
   * @throws EmptyBufferException if the buffer is empty
   */
  public Object removeMessage() throws EmptyBufferException
  {
    synchronized (queue)
    {
      if (queue.isEmpty())
        throw new EmptyBufferException();
      return queue.remove(0);
    }
  }

  /**
   * Returns true if the buffer is empty.
   * 
   * @return true if the buffer is empty.
   */
  public synchronized boolean isEmpty()
  {
    return queue.isEmpty();
  }

}