/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel;

import java.io.Serializable;
import java.util.ArrayList;

import org.objectweb.tribe.channel.tcp.TcpChannelPool;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.EmptyBufferException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.messages.GroupMessage;

/**
 * This class defines a ReliableGroupChannel
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class ReliableGroupChannel
{
  protected AbstractChannelPool channelPool;
  protected ReceiveBuffer       incomingBuffer;
  protected Group               currentGroup;
  protected Member              me = null;

  /**
   * Creates a new <code>ReliableGroupChannel</code> object without group
   * membership service.
   */
  public ReliableGroupChannel()
  {
    channelPool = TcpChannelPool.getChannelPool();
  }

  /**
   * Join the group that has the given group identifier.
   * 
   * @param gid the identifier of the group to join
   * @throws AlreadyMemberException if we are already member of the group
   * @throws NotConnectedException if the channel is not connected
   * @throws ChannelException if an error is reported by the channel
   */
  public void join(Group g) throws AlreadyMemberException, ChannelException,
      NotConnectedException
  {
    if ((currentGroup != null) && currentGroup.equals(g))
      throw new AlreadyMemberException();
    // Quit the previous group first if needed
    if (currentGroup != null)
      try
      {
        quit();
      }
      catch (Exception ignore)
      {
      }
    currentGroup = g;
    incomingBuffer = new ReceiveBuffer(currentGroup.getGroupIdentifier());
    channelPool.registerReceiveBuffer(incomingBuffer);
  }

  /**
   * Quit the current group without closing the channel.
   * 
   * @throws ChannelException if a communication error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   */
  public void quit() throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    channelPool.unregisterReceiveBuffer(incomingBuffer);
    currentGroup = null;
  }

  /**
   * Returns the current group we are connected to or null if we are not
   * connected to any group.
   * 
   * @return Returns the current group or null.
   */
  public Group getCurrentGroup()
  {
    return currentGroup;
  }

  /**
   * Sends the given message to all members of the group.
   * 
   * @param msg message to send
   * @return an <code>ArrayList</code> of Members who failed, or null if all
   *         members received successfully the message.
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   */
  public ArrayList send(Serializable msg) throws ChannelException,
      NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    return send(msg, currentGroup.getGroupIdentifier(), currentGroup
        .getMembers());
  }

  /**
   * Sends a message to a subset of group members.
   * 
   * @param msg message to send
   * @param members <code>ArrayList</code> of <code>Member</code> that are
   *          all part of the group
   * @return an <code>ArrayList</code> of Members who failed, or null if all
   *         members received successfully the message.
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   */
  public ArrayList send(Serializable msg, ArrayList members)
      throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    return send(msg, currentGroup.getGroupIdentifier(), members);
  }

  /**
   * Sends a message to members of group gid without being necessary member of
   * the group.
   * 
   * @param msg message to send
   * @param members <code>ArrayList</code> of <code>Member</code> that are
   *          all part of the group
   * @return an <code>ArrayList</code> of Members who failed, or null if all
   *         members received successfully the message.
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   */
  public ArrayList send(Serializable msg, GroupIdentifier gid, ArrayList members)
      throws ChannelException, NotConnectedException
  {
    return channelPool.send(new GroupMessage(msg, gid), members);
  }

  /**
   * Receives a new message from the channel.
   * 
   * @return the new message
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   */
  public Serializable receive() throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    try
    {
      GroupMessage groupMessage = (GroupMessage) incomingBuffer.getMessage();
      if (groupMessage == null)
        return null;
      else
        return groupMessage.getMessage();
    }
    catch (EmptyBufferException e)
    {
      throw new ChannelException("Error while retrieving message from buffer",
          e);
    }
  }

  /**
   * Closes the channel (also quits the current group if any).
   * 
   * @throws ChannelException if a network error occurs
   * @throws NotConnectedException if the channel is not connected
   */
  public void close() throws ChannelException, NotConnectedException
  {
    if (currentGroup != null)
      try
      {
        this.quit();
      }
      catch (Exception ignore)
      {
      }
  }

  /**
   * Returns the local membership value (Member object representing this channel
   * in the group).
   * 
   * @return Returns the local membership (me).
   */
  public Member getLocalMembership()
  {
    return me;
  }

  /**
   * Sets the membership value (me as a Member of the group).
   * 
   * @param membership The me to set.
   * @throws NotConnectedException if the channel is not currently connected to
   *           any group
   */
  protected void setMembership(Member membership) throws NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    this.me = membership;
  }
}