/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel;

import java.io.Serializable;
import java.util.ArrayList;

import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.gms.GroupMembershipService;

/**
 * This class defines a ReliableGroupChannel relying on a Group Membership
 * Service (GMS).
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class ReliableGroupChannelWithGms extends ReliableGroupChannel
{
  protected GroupMembershipService gms = null;

  /**
   * Creates a new <code>ReliableGroupChannelWithGms</code> using the provided
   * group membership service.
   * 
   * @param gms the GMS service to use
   */
  public ReliableGroupChannelWithGms(GroupMembershipService gms)
  {
    super();
    this.gms = gms;
  }

  /**
   * Join the group that has the given group identifier.
   * 
   * @param gid the identifier of the group to join
   * @throws AlreadyMemberException if we are already member of the group
   * @throws NotConnectedException if the channel is not connected
   * @throws ChannelException if an error is reported by the channel
   */
  public void join(GroupIdentifier gid) throws AlreadyMemberException,
      ChannelException, NotConnectedException
  {
    Group g = getGroup(gid);
    if (g == null)
      g = new Group(gid);
    super.join(g);
    me = gms.join(this, gid);
    currentGroup = getGroup(gid);
  }

  /**
   * Quit the current group.
   * 
   * @throws ChannelException if a communication error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   */
  public void quit() throws ChannelException, NotConnectedException
  {
    if (currentGroup == null)
      throw new NotConnectedException();
    try
    {
      gms.quit(this, currentGroup.getGroupIdentifier());
    }
    catch (ChannelException ignore)
    {
    }
    super.quit();
    me = null;
  }

  /**
   * Get the group (including its members) corresponding to the given gid. If
   * the GMS was never asked to join the given group, then null is returned.
   * 
   * @param gid the group identifier
   * @return the <code>Group</code> corresponding to this gid.
   */
  public Group getGroup(GroupIdentifier gid)
  {
    return gms.getGroup(gid);
  }

  /**
   * Sends a message to members of group gid without being necessary member of
   * the group.
   * 
   * @param msg message to send
   * @param members <code>ArrayList</code> of <code>Member</code> that are
   *          all part of the group
   * @return an <code>ArrayList</code> of Members who failed, or null if all
   *         members received successfully the message.
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if the channel is not connected to any group
   */
  public ArrayList send(Serializable msg, GroupIdentifier gid, ArrayList members)
      throws ChannelException, NotConnectedException
  {
    ArrayList failed = super.send(msg, gid, members);
    if (failed != null)
    { // Removed all failed members from group
      for (int i = 0; i < failed.size(); i++)
        gms.quitMember((Member) failed.get(i), gid);
    }
    return failed;
  }

}