/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel.tcp;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;

import org.objectweb.tribe.channel.AbstractReliableFifoChannel;
import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.messages.ChannelMessage;

/**
 * This class defines a TcpChannel
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class TcpChannel extends AbstractReliableFifoChannel
{
  // underlying TCP socket
  private Socket           socket;
  private IpAddress        destinationAddress = null;
  private IpAddress        sourceAddress      = null;
  private DataInputStream  inStream;
  private DataOutputStream outStream;
  private boolean          isClosed;

  /**
   * Creates a new <code>TcpChannel</code> using the local host IP and a port
   * number choosen by the system.
   * 
   * @throws IOException if an error occurs
   */
  public TcpChannel() throws IOException
  {
    socket = new Socket();
    sourceAddress = new IpAddress(socket.getLocalAddress(), socket
        .getLocalPort());
    initializeStreams();
  }

  /**
   * Creates a new <code>TcpChannel</code> binding the source endpoint to the
   * given address.
   * 
   * @param sourceAddress address to bind the socket to
   * @throws IOException if an error occurs
   */
  public TcpChannel(IpAddress sourceAddress) throws IOException
  {
    socket = new Socket(sourceAddress.getAddress(), sourceAddress.getPort(),
        true);
    this.sourceAddress = sourceAddress;
    initializeStreams();
  }

  /**
   * Creates a new <code>TcpChannel</code> from an existing socket.
   * 
   * @param socket a TCP socket
   * @throws IOException if an error occurs
   */
  public TcpChannel(Socket socket) throws IOException
  {
    this.socket = socket;
    sourceAddress = new IpAddress(socket.getLocalAddress(), socket
        .getLocalPort());
    destinationAddress = new IpAddress(socket.getInetAddress(), socket
        .getPort());
    initializeStreams();
  }

  /**
   * Disable the Nagle algorithm for better latency and create object input and
   * output streams.
   * 
   * @throws IOException if an error occurs
   */
  private void initializeStreams() throws IOException
  {
    socket.setTcpNoDelay(false);
    isClosed = false;
  }

  /**
   * @see org.objectweb.tribe.channel.AbstractReliableFifoChannel#send(org.objectweb.tribe.messages.ChannelMessage)
   */
  public synchronized void send(ChannelMessage msg) throws ChannelException,
      NotConnectedException
  {
    // Sanity checks
    if (isClosed || (socket == null))
      throw new NotConnectedException();
    if (outStream == null)
    { // Create the stream if it does not exist
      try
      {
        outStream = new DataOutputStream(new BufferedOutputStream(socket
            .getOutputStream()));
      }
      catch (IOException e1)
      {
        throw new ChannelException("Unable to create output stream", e1);
      }
    }

    // Send the message
    try
    {
      byte[] msgInBytes = msg.getByteArray();
      // Note that we do not send the message as an Object but directly in its
      // serialized form to prevent multiple serializations of the object.
      outStream.writeInt(msgInBytes.length);
      outStream.write(msgInBytes);
      outStream.flush();
    }
    catch (IOException e)
    {
      throw new ChannelException("Error while sending message on socket", e);
    }
  }

  /**
   * @see org.objectweb.tribe.channel.AbstractReliableFifoChannel#close()
   */
  public void close() throws ChannelException
  {
    if (isClosed)
      return;
    try
    {
      socket.close();
    }
    catch (IOException e)
    {
      throw new ChannelException("Error while closing the socket", e);
    }
    finally
    {
      isClosed = true;
    }
  }

  /**
   * @see org.objectweb.tribe.channel.AbstractReliableFifoChannel#connect(org.objectweb.tribe.common.Address)
   */
  public void connect(Address destination) throws ChannelException
  {
    if (!(destination instanceof IpAddress))
      throw new ChannelException("TCP Channels require IP addresses.");
    destinationAddress = (IpAddress) destination;
    try
    {
      socket.connect(new InetSocketAddress(destinationAddress.getAddress(),
          destinationAddress.getPort()));
    }
    catch (IOException e)
    {
      inStream = null;
      outStream = null;
      throw new ChannelException("Error while connecting the socket", e);
    }
  }

  /**
   * @see org.objectweb.tribe.channel.AbstractReliableFifoChannel#getDestinationAddress()
   */
  public Address getDestinationAddress()
  {
    return destinationAddress;
  }

  /**
   * @see org.objectweb.tribe.channel.AbstractReliableFifoChannel#getSourceAddress()
   */
  public Address getSourceAddress()
  {
    return sourceAddress;
  }

  /**
   * Returns the inStream value.
   * 
   * @return Returns an input stream on the underlying socket or null if no
   *         input stream can be created.
   */
  protected DataInputStream getInStream()
  {
    if (socket == null)
      return null;
    if (inStream == null)
      try
      {
        inStream = new DataInputStream(new BufferedInputStream(socket
            .getInputStream()));
      }
      catch (IOException e)
      {
        return null;
      }

    return inStream;
  }
}