/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel.tcp;

import java.io.IOException;
import java.util.HashMap;

import org.objectweb.tribe.channel.AbstractChannelPool;
import org.objectweb.tribe.channel.AbstractReliableFifoChannel;
import org.objectweb.tribe.channel.AbstractServerChannel;
import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.ChannelException;

/**
 * This class defines a TcpChannelPool
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class TcpChannelPool extends AbstractChannelPool
{

  private static TcpChannelPool pool   = new TcpChannelPool();
  private HashMap               readerThreads;
  private HashMap               accepterThreads;
  private static Trace          logger = Trace
                                           .getLogger("org.objectweb.tribe.channel");

  /**
   * Creates a new <code>TcpChannelPool</code> object
   */
  public TcpChannelPool()
  {
    super();
    readerThreads = new HashMap();
    accepterThreads = new HashMap();
  }

  /**
   * @see org.objectweb.tribe.serverSocket.pool.AbstractChannelPool#getChannel(org.objectweb.tribe.common.Address)
   */
  public AbstractReliableFifoChannel getChannel(Address destination)
      throws ChannelException
  {
    synchronized (channels)
    {
      AbstractReliableFifoChannel channel = (AbstractReliableFifoChannel) channels
          .get(destination);
      if (channel == null)
      {
        try
        {
          if (logger.isDebugEnabled())
            logger.debug("Getting new channel for " + destination);
          channel = new TcpChannel();
          channel.connect(destination);
          channels.put(destination, channel);
          TcpReaderThread thread = new TcpReaderThread((TcpChannel) channel,
              keyBuffers);
          thread.start();
          synchronized (readerThreads)
          {
            readerThreads.put(channel, thread);
          }
        }
        catch (IOException e)
        {
          throw new ChannelException("Failed to create a new serverSocket to "
              + destination, e);
        }
      }
      return channel;
    }
  }

  /**
   * @see org.objectweb.tribe.channel.AbstractChannelPool#getServerChannel(org.objectweb.tribe.common.Address)
   */
  public AbstractServerChannel getServerChannel(Address serverAddress)
      throws ChannelException
  {
    synchronized (serverChannels)
    {
      AbstractServerChannel channel = (AbstractServerChannel) serverChannels
          .get(serverAddress);
      if (channel == null)
      {
        try
        {
          if (logger.isDebugEnabled())
            logger.debug("Getting new server channel for " + serverAddress);
          int port = ((IpAddress) serverAddress).getPort();
          if (port == 0)
            channel = new TcpServerChannel(port);
          else
          {
            channel = new TcpServerChannel();
            channel.bind(serverAddress);
          }
          serverChannels.put(serverAddress, channel);
          TcpServerAccepterThread thread = new TcpServerAccepterThread(channel,
              channels, readerThreads, keyBuffers);
          thread.start();
          synchronized (accepterThreads)
          {
            accepterThreads.put(channel, thread);
          }
        }
        catch (IOException e)
        {
          throw new ChannelException(
              "Failed to create a new server serverSocket on " + serverAddress,
              e);
        }
      }
      return channel;
    }
  }

  /**
   * Returns a static instance of the serverSocket pool.
   * 
   * @return a <code>TcpChannelPool</code>
   */
  public static AbstractChannelPool getChannelPool()
  {
    return pool;
  }

  /**
   * Terminates the TCP reader thread as well.
   * 
   * @see org.objectweb.tribe.channel.AbstractChannelPool#removeChannelFromPool(org.objectweb.tribe.channel.AbstractReliableFifoChannel)
   */
  public boolean removeChannelFromPool(AbstractReliableFifoChannel channel)
  {
    TcpReaderThread thread;
    synchronized (readerThreads)
    {
      thread = (TcpReaderThread) readerThreads.remove(channel);
    }
    if (thread != null)
      thread.kill();
    return super.removeChannelFromPool(channel);
  }

  /**
   * Terminates the TCP accepter thread as well.
   * 
   * @see org.objectweb.tribe.channel.AbstractChannelPool#removeServerChannelFromPool(org.objectweb.tribe.channel.AbstractServerChannel)
   */
  public boolean removeServerChannelFromPool(AbstractServerChannel channel)
  {
    TcpServerAccepterThread thread;
    synchronized (accepterThreads)
    {
      thread = (TcpServerAccepterThread) accepterThreads.remove(channel);
    }
    if (thread != null)
      thread.kill();
    return super.removeServerChannelFromPool(channel);
  }
}