/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel.tcp;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.HashMap;

import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NoReceiverException;
import org.objectweb.tribe.messages.ChannelMessage;

/**
 * This class defines a TcpReaderThread.
 * <p>
 * This thread reads messages from a TCP connection and store them in the
 * appropriate receive buffers.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class TcpReaderThread extends Thread
{
  boolean              isKilled = false;
  private TcpChannel   channel;
  private HashMap      keyBuffers;

  private static Trace logger   = Trace
                                    .getLogger("org.objectweb.tribe.channel");

  /**
   * Creates a new <code>TcpReaderThread</code> object
   * 
   * @param logger logger for debug
   * @param objectInputStream socket input stream to read from
   * @param keyBuffers list of pool receive buffers
   */
  public TcpReaderThread(TcpChannel channel, HashMap keyBuffers)
  {
    super("TcpReaderThread");
    this.channel = channel;
    this.keyBuffers = keyBuffers;
  }

  /**
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    if (logger.isDebugEnabled())
      logger.debug("TcpReaderThread started on " + channel.getSourceAddress()
          + "->" + channel.getDestinationAddress());

    DataInputStream inStream = channel.getInStream();

    while (!isKilled)
    {
      try
      {
        // Protocol is:
        // 1. message size (as an int)
        // 2. read an array of bytes of the given size and convert it to an
        // object.
        int size = inStream.readInt(); // number of bytes to read
        int totalRead = 0; // number of bytes read
        byte[] buf = new byte[size];
        do
        {
          int read = inStream.read(buf, totalRead, size);
          totalRead += read;
          size -= read;
        }
        while (size > 0);

        ChannelMessage msg = (ChannelMessage) new ObjectInputStream(
            new ByteArrayInputStream(buf)).readObject();
        if (logger.isDebugEnabled())
          logger.debug("TcpReaderThread received message: " + msg);
        msg.deliverMessage(keyBuffers);
      }
      catch (RuntimeException e)
      {
        if (logger.isDebugEnabled())
          logger
              .debug("TcpReaderThread: Error while receiving message, terminating thread and channel ("
                  + e + ")");
        try
        {
          channel.close();
        }
        catch (ChannelException ignore)
        {
        }
        isKilled = true;
      }
      catch (IOException e)
      {
        if (logger.isDebugEnabled())
          logger
              .debug("TcpReaderThread: Error while receiving message, terminating thread and channel ("
                  + e + ")");
        try
        {
          channel.close();
        }
        catch (ChannelException ignore)
        {
        }
        isKilled = true;
      }
      catch (ClassNotFoundException e)
      {
        logger.error("TcpReaderThread: Error while unmarshalling message", e);
      }
      catch (NoReceiverException e)
      {
        logger.info("TcpReaderThread: Error while delivering message", e);
      }
    }

    if (logger.isDebugEnabled())
      logger.debug("TcpReaderThread terminated.");
  }

  /**
   * Terminates this thread
   */
  public void kill()
  {
    isKilled = true;
    // Close the stream to wake up the thread
    try
    {
      channel.getInStream().close();
    }
    catch (IOException ignore)
    {
    }
    try
    {
      channel.close();
    }
    catch (ChannelException ignore)
    {
    }
    this.interrupt();
  }

}