/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel.tcp;

import java.util.HashMap;

import org.objectweb.tribe.channel.AbstractReliableFifoChannel;
import org.objectweb.tribe.channel.AbstractServerChannel;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.ChannelException;

/**
 * This class defines a TcpServerAccepterThread.
 * <p>
 * It simply accepts connections and put them in the connection pool.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class TcpServerAccepterThread extends Thread
{
  private AbstractServerChannel serverSocket;
  // IP address -> channel
  private HashMap               channels;
  // channel -> TCP reader thread
  private HashMap               readerThreads;
  private HashMap               keyBuffers;
  private boolean               isKilled = false;

  private static Trace          logger   = Trace
                                             .getLogger("org.objectweb.tribe.channel");

  /**
   * Creates a new <code>TcpServerAccepterThread</code> object
   * 
   * @param serverSocket serverSocket to read from
   * @param channels TcpPool channel hashmap to update when a new connection is
   *          created.
   */
  public TcpServerAccepterThread(AbstractServerChannel serverSocket,
      HashMap channels, HashMap readerThreads, HashMap keyBuffers)
  {
    super("TcpServerAccepterThread");
    this.serverSocket = serverSocket;
    this.channels = channels;
    this.keyBuffers = keyBuffers;
    this.readerThreads = readerThreads;
  }

  /**
   * Accept new connections and put them in the pool
   */
  public void run()
  {
    if (logger.isDebugEnabled())
      logger.debug("TcpServerAccepterThread started");

    while (!isKilled)
    {
      try
      {
        AbstractReliableFifoChannel socket = serverSocket.accept();
        IpAddress ip = (IpAddress) socket.getSourceAddress();
        synchronized (channels)
        {
          if (logger.isDebugEnabled())
            logger.debug("Accepting new connection from " + ip);
          // TODO: Handle duplicate entries?
          channels.put(ip, socket);
          TcpReaderThread thread = new TcpReaderThread((TcpChannel) socket,
              keyBuffers);
          thread.start();
          readerThreads.put(socket, thread);
        }
      }
      catch (ChannelException e)
      {
        logger.info("Channel error, exiting TcpServerAccepterThread", e);
        isKilled = true;
      }
    }

    if (logger.isDebugEnabled())
      logger.debug("TcpServerAccepterThread terminated.");
  }

  /**
   * Terminates this TcpServerAccepterThread.
   */
  public void kill()
  {
    isKilled = true;
    try
    {
      serverSocket.close();
    }
    catch (ChannelException ignore)
    {
    }
  }
}