/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel.tcp;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;

import org.objectweb.tribe.channel.AbstractReliableFifoChannel;
import org.objectweb.tribe.channel.AbstractServerChannel;
import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;

/**
 * This class defines a TcpServerChannel
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class TcpServerChannel extends AbstractServerChannel
{

  private ServerSocket socket; // underlying TCP server socket

  /**
   * Creates a new unbound <code>TcpServerChannel</code>
   * 
   * @throws IOException if an error occurs
   */
  public TcpServerChannel() throws IOException
  {
    socket = new ServerSocket();
  }

  /**
   * Creates a new <code>TcpServerChannel</code> bound to the specified port.
   * If port is 0, then the socket is bound to an available port chosen by the
   * system.
   * 
   * @param port bind the socket to the given port (0 let the system choose an
   *          available port).
   * @throws IOException if an error occurs
   */
  public TcpServerChannel(int port) throws IOException
  {
    socket = new ServerSocket(port);
  }

  /**
   * @see org.objectweb.tribe.serverSocket.AbstractServerChannel#bind(org.objectweb.tribe.common.Address)
   */
  public void bind(Address source) throws ChannelException
  {
    if (!(source instanceof IpAddress))
      throw new ChannelException("TCP Channels require IP addresses.");
    IpAddress ip = (IpAddress) source;
    try
    {
      socket.bind(new InetSocketAddress(ip.getAddress(), ip.getPort()));
    }
    catch (IOException e)
    {
      throw new ChannelException("Error while binding serverSocket on "
          + source, e);
    }
  }

  /**
   * @see org.objectweb.tribe.channel.AbstractServerChannel#getBindAddress()
   */
  public Address getBindAddress() throws NotConnectedException
  {
    if (socket == null)
      throw new NotConnectedException();
    return new IpAddress(socket.getInetAddress(), socket.getLocalPort());
  }

  /**
   * @see org.objectweb.tribe.serverSocket.AbstractServerChannel#accept()
   */
  public AbstractReliableFifoChannel accept() throws ChannelException
  {
    try
    {
      return new TcpChannel(socket.accept());
    }
    catch (Exception e)
    {
      throw new ChannelException("Error while accepting new connection", e);
    }
  }

  /**
   * @see org.objectweb.tribe.serverSocket.AbstractServerChannel#close()
   */
  public void close() throws ChannelException
  {
    try
    {
      socket.close();
    }
    catch (IOException e)
    {
      throw new ChannelException("Error while closing the serverSocket", e);
    }
    finally
    {
      socket = null;
    }
  }

}