/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): Nicolas Modrzyk.
 */

package org.objectweb.tribe.common;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class defines a Group.
 * <p>
 * Note that the member list is always sorted in ascending order of member
 * addresses.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @author <a href="mailto:Nicolas.Modrzyk@inrialpes.fr">Nicolas Modrzyk </a>
 * @version 1.0
 */
public class Group implements Serializable
{
  private GroupIdentifier groupIdentifier;
  private ArrayList       members;

  /**
   * Creates a new <code>Group</code> with the given group identifier and no
   * members.
   */
  public Group(GroupIdentifier gid)
  {
    groupIdentifier = gid;
    members = new ArrayList();
  }

  /**
   * Get the identifier of this group.
   * 
   * @return a <code>GroupIdentifier</code>.
   */
  public GroupIdentifier getGroupIdentifier()
  {
    return groupIdentifier;
  }

  /**
   * The list of members in the group. This is not a copy of the member list and
   * so any modification to the list will implicitely modify the group
   * composition.
   * 
   * @return an <code>ArrayList</code> of <code>Member</code>.
   */
  public ArrayList getMembers()
  {
    return members;
  }

  /**
   * Adds a member to the group. This function does not look for duplicate group
   * members.
   * 
   * @param m the new group member
   */
  public void addMember(Member m)
  {
    Address addr = m.getAddress();
    synchronized (members)
    {
      int size = members.size();
      for (int i = 0; i < size; i++)
      {
        Member member = (Member) members.get(i);
        if (addr.compareTo(member.getAddress()) < 0)
        {
          members.add(i, m);
          return;
        }
      }
      // This is the last element
      members.add(m);
    }
  }

  /**
   * Remove a member from this group.
   * 
   * @param member <code>Member</code> object.
   * @return true if the member was is the group, false otherwise.
   */
  public boolean removeMember(Member member)
  {
    synchronized (members)
    {
      return members.remove(member);
    }
  }

  /**
   * Tests if a member belongs to this group.
   * 
   * @param member the member to look for
   * @return true if the member is in the group, false otherwise
   */
  public boolean hasMember(Member member)
  {
    synchronized (members)
    {
      return members.contains(member);
    }
  }

  /**
   * Merge a group definition with another one
   * 
   * @param other the other group to merge member with.
   * @return true if the group composition has been modified, false otherwise
   */
  public boolean merge(Group other)
  {
    boolean hasBeenModified = false;
    synchronized (members)
    {
      for (Iterator iter = other.getMembers().iterator(); iter.hasNext();)
      {
        Member member = (Member) iter.next();
        if (!members.contains(member))
        {
          hasBeenModified = true;
          addMember(member);
        }
      }
    }
    return hasBeenModified;
  }

  /**
   * Return the list of members as a String.
   * 
   * @return String containing member list
   */
  public String getStringMembers()
  {
    StringBuffer sb = new StringBuffer();
    synchronized (members)
    {
      for (Iterator iter = getMembers().iterator(); iter.hasNext();)
      {
        Member member = (Member) iter.next();
        sb.append(member.toString() + " ");
      }
    }
    return sb.toString();
  }

}