/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Nicolas Modrzyk.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.faultdetection;

import java.util.ArrayList;

import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.Member;

/**
 * This class defines a FaultDetectionService
 * 
 * @author <a href="mailto:Nicolas.Modrzyk@inrialpes.fr">Nicolas Modrzyk </a>
 * @version 1.0
 */
public interface FaultDetectionService
{
  /**
   * Default refresh value
   */
  static final int DEFAULT_REFREST_RATE = 1000;
  
  /**
   * The default port for fault detection check
   */
  static final int DEFAULT_LISTENING_PORT = 6456;
  
  /**
   * The default emitting port for fault detection check
   */
  static final int DEFAULT_EMITTING_PORT = 6457;

  /**
   * The intervals to check for new or lost members
   * 
   * @param delay <code>long</code> value in milliseconds
   */
  void setRefreshRate(long delay);

  /**
   * Return the value used for refresh rate. If no value has been set, a default
   * value is used.
   * 
   * @return <code>long</code> value in milliseconds
   * @see DEFAULT_REFREST_RATE
   */
  long getRefreshRate();

  /**
   * Add a group to analyse
   * 
   * @param group <code>Group</code> object that contains the members to
   *                    listen to
   */
  void addGroup(Group group);
  
  /** 
   * Add a member to the ping service
   * 
   * @param member <code>Member</code> to listen to
   */
  void addMember(Member member);
  
  /**
   * Remove this member from the service
   * 
   * @param member <code>Member</code> to stop listening to
   */
  void removeMember(Member member);

  /**
   * Stop detecting failure for this group
   * 
   * @param group <code>Group</code> object that contains the members to
   *                    listen to
   * @return group if the group was listened to before, null if no changes has
   *                 been made.
   */
  Group removeGroup(Group group);

  /**
   * Access all the groups the service is listening to. This is the original
   * list of groups so any changes to this list will effectivel change the
   * groups we are analysing
   * 
   * @return an <code>ArrayList</code> of <code>Group</code>
   */
  ArrayList getGroups();
  
  /**
   * Add a listener of event 
   * 
   * @param listener <code>FaultDetectionListener</code> to send notification to
   */
  void addListener(FaultDetectionListener listener);
  
  /**
   * Tell the fault service to no more send event to the listener 
   * 
   * @param listener <code>FaultDetectionListener</code> to send notification to
   */
  void removeListener(FaultDetectionListener listener);
}