/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Nicolas Modrzyk.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.faultdetection;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Hashtable;

import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.common.Service;

/**
 * This class defines a PingFaultDetectionService. This will detect missing
 * members by pinging them using regular emission of packets.
 * 
 * @author <a href="mailto:Nicolas.Modrzyk@inrialpes.fr">Nicolas Modrzyk </a>
 * @version 1.0
 */
public class PingFaultDetectionService
    implements
      FaultDetectionService,
      Service
{
  /** Contains a list of group that the fault detection service is listening to */
  private ArrayList         groups;
  /** The ping manager thread, that will manager ping on members */
  private PingManagerThread pingThread;
  /** Hashtable of pong threads that will listen and respond to ping thread */
  private Hashtable         pongThreads;
  /** The local address */
  private IpAddress         localAddress;

  /** The port that will be used to emit ping */
  private int               emittingPort;
  /** The port that will be used to listen to ping */
  private int               listeningPort;

  /**
   * Creates a new <code>PingFaultDetectionService</code> object
   * 
   * @throws UnknownHostException
   */
  public PingFaultDetectionService() throws UnknownHostException
  {
    this(FaultDetectionService.DEFAULT_EMITTING_PORT,
        FaultDetectionService.DEFAULT_LISTENING_PORT);
  }

  /**
   * Creates a new <code>PingFaultDetectionService</code> object
   * 
   * @param emittingPort
   * @param listeningPort
   * @throws UnknownHostException
   */
  public PingFaultDetectionService(int emittingPort, int listeningPort) throws UnknownHostException
  {
    this.emittingPort = emittingPort;
    this.listeningPort = listeningPort;
    groups = new ArrayList(10);
    pongThreads = new Hashtable();
    initThreads();
  }

  private void initThreads() throws UnknownHostException
  {
    pingThread = new PingManagerThread(localAddress,
        FaultDetectionService.DEFAULT_REFREST_RATE);
    localAddress = new IpAddress(InetAddress.getLocalHost(), emittingPort);
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#setRefreshRate(long)
   */
  public void setRefreshRate(long delay)
  {
    pingThread.setWaitInterval(delay);
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#getRefreshRate()
   */
  public long getRefreshRate()
  {
    return pingThread.getWaitInterval();
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#addGroup(org.objectweb.tribe.common.Group)
   */
  public void addGroup(Group group)
  {
    // add group to service
    synchronized (groups)
    {
      groups.add(group);
    }
    // addentries in ping thread
    ArrayList members = group.getMembers();
    for (int i = 0; i < members.size(); i++)
      addMember((Member) members.get(i));
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#addMember(org.objectweb.tribe.common.Member)
   */
  public void addMember(Member member)
  {
    // Get the address of the member
    IpAddress address = (IpAddress) member.getAddress();
    // Create a new address with the same host the and the listening port for
    // fault detection
    IpAddress pongAddress = new IpAddress(address.getAddress(), listeningPort);
    // Tell the ping thread to ping this address
    pingThread.addPingHook(address);
    // Create and start a pong thread that will response to ping from this
    // address
    UDPPongThread pong = new UDPPongThread(localAddress, pongAddress);
    new Thread(pong).start();
    // Put the pong thread in the hashtable to remove it in the future
    pongThreads.put(address, pong);
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#removeMember(org.objectweb.tribe.common.Member)
   */
  public void removeMember(Member member)
  {
    Address address = member.getAddress();
    // Stop pinging this address
    pingThread.removePingHook(address);
    // Remove the pong thread from the hashtable
    UDPPongThread thread = (UDPPongThread) pongThreads.remove(address);
    // Stop answering to ping from this address
    if (thread != null)
      thread.setQuit(true);
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#removeGroup(org.objectweb.tribe.common.Group)
   */
  public Group removeGroup(Group group)
  {
    if (group == null)
      return null;
    boolean check = false;
    synchronized (groups)
    {
      check = groups.contains(group) && groups.remove(group);
    }
    if (check)
    {
      // Remove addresses from ping thread
      ArrayList members = group.getMembers();
      for (int i = 0; i < members.size(); i++)
        removeMember((Member) members.get(i));
      // Return group
      return group;
    }
    else
      return null;
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#getGroups()
   */
  public ArrayList getGroups()
  {
    return groups;
  }

  /**
   * @see org.objectweb.tribe.common.Service#start()
   */
  public void start()
  {
    // Start ping thread
    new Thread(pingThread).start();
  }

  /**
   * @see org.objectweb.tribe.common.Service#stop()
   */
  public void stop()
  {
    // Ask ping thread to stop
    pingThread.setStop(true);
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#addListener(org.objectweb.tribe.faultdetection.FaultDetectionListener)
   */
  public void addListener(FaultDetectionListener listener)
  {
    // Add listener to ping thread
    pingThread.addListener(listener);
  }

  /**
   * @see org.objectweb.tribe.faultdetection.FaultDetectionService#removeListener(org.objectweb.tribe.faultdetection.FaultDetectionListener)
   */
  public void removeListener(FaultDetectionListener listener)
  {
    // Remove listener from ping thread
    pingThread.removeListener(listener);
  }
}