/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Nicolas Modrzyk.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.faultdetection;

import java.net.SocketException;
import java.util.ArrayList;
import java.util.Hashtable;

import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.IpAddress;

/**
 * This class defines a PingManagerThread. A ping thread will periodically send
 * ping to addresses and send events to registered listeners.
 * 
 * @author <a href="mailto:Nicolas.Modrzyk@inrialpes.fr">Nicolas Modrzyk </a>
 * @version 1.0
 */
public class PingManagerThread implements Runnable
{

  ArrayList addresses;
  ArrayList listeners;
  Hashtable sockets;
  long      waitInterval;
  boolean   stop;
  IpAddress localAddress;

  /**
   * Creates a new <code>PingManagerThread</code> object
   * 
   * @param localAddress
   * @param waitInterval the time between two ping
   */
  public PingManagerThread(IpAddress localAddress, long waitInterval)
  {
    super();
    addresses = new ArrayList();
    listeners = new ArrayList();
    sockets = new Hashtable();
    this.waitInterval = waitInterval;
    this.localAddress = localAddress;
    stop = false;
  }

  /**
   * Tell the thread to stop as soon as possible
   * 
   * @param stop The stop to set.
   */
  public void setStop(boolean stop)
  {
    this.stop = stop;
  }

  /**
   * Returns the waitInterval value.
   * 
   * @return Returns the waitInterval.
   */
  public long getWaitInterval()
  {
    return waitInterval;
  }

  /**
   * Sets the waitInterval value.
   * 
   * @param waitInterval The waitInterval to set.
   */
  public void setWaitInterval(long waitInterval)
  {
    this.waitInterval = waitInterval;
  }

  /**
   * Tell the thread it shoudl start pinging a new address
   * 
   * @param address <code>Address</code> to ping
   */
  public void addPingHook(Address address)
  {
    synchronized (addresses)
    {
      addresses.add(address);
    }
  }

  /**
   * Tell the thread to stop pinging an address
   * 
   * @param address <code>Address</code> to stop pinging
   */
  public void removePingHook(Address address)
  {
    synchronized (addresses)
    {
      addresses.remove(address);
    }
  }

  /**
   * Register a new ping listener, that will receive ping events
   * 
   * @param listener <code>FaultDetectionListener</code> object that will then
   *                    receive events and results of ping action
   */
  public void addListener(FaultDetectionListener listener)
  {
    synchronized (listeners)
    {
      listeners.add(listener);
    }
  }

  /**
   * Unregister a ping listener
   * 
   * @param listener <code>FaultDetectionListener</code> will not receive
   *                    further events
   */
  public void removeListener(FaultDetectionListener listener)
  {
    synchronized (listeners)
    {
      listeners.remove(listener);
    }
  }

  /**
   * Execute a ping on an <code>Address</code>. This does not collect the
   * result
   * 
   * @param address <code>Address</code> address object to check for
   */
  private void doPing(IpAddress address)
  {
    Thread t = (Thread) sockets.get(address);
    if (t == null)
    {
      try
      {
        t = new UDPPingThread(localAddress, address, (int) waitInterval);
      }
      catch (SocketException e)
      {
        notifyListenerOfEvent(FaultDetectionListener.EVENT_ERROR_LOCAL_ADDRESS,
            localAddress);
        return;
      }
    }
    t.start();
  }

  /**
   * Get the result and call back events on listeners 
   * 
   * @param address <code>IpAddress</code> of the result to get
   */
  public void getResult(IpAddress address)
  {
    UDPPingThread t = (UDPPingThread) sockets.get(address);
    if (!t.isTargetAlive())
    {
      notifyListenerOfEvent(FaultDetectionListener.THREAD_SUSPECTED, address);
    }
    else
    {
      notifyListenerOfEvent(FaultDetectionListener.THREAD_ALIVE, address);
    }
  }

  /**
   * Notify listeners of an event on the specified address
   * 
   * @see FaultDetectionService
   * @param event an int representing the event value
   * @param address the address where the event occured
   */
  public void notifyListenerOfEvent(int event, IpAddress address)
  {
    switch (event)
    {
      case FaultDetectionListener.THREAD_ALIVE :
        synchronized (listeners)
        {
          for (int i = 0; i < listeners.size(); i++)
            ((FaultDetectionListener) listeners.get(i)).alive(address);
        }
        break;
      case FaultDetectionListener.THREAD_SUSPECTED :
        synchronized (listeners)
        {
          for (int i = 0; i < listeners.size(); i++)
            ((FaultDetectionListener) listeners.get(i)).noResponse(address);
        }
        break;
      default :
        synchronized (listeners)
        {
          for (int i = 0; i < listeners.size(); i++)
            ((FaultDetectionListener) listeners.get(i)).pingServiceEvent(event);
        }
        break;
    }
  }

  /**
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    while (!stop)
    {

      // UDPPingThread all addresses
      synchronized (addresses)
      {
        for (int i = 0; i < addresses.size(); i++)
          doPing(((IpAddress) addresses.get(i)));
      }

      try
      {
        // Wait till next ping time
        synchronized (this)
        {
          wait(waitInterval);
        }
      }
      catch (InterruptedException e)
      {
        // Notify listeners
        notifyListenerOfEvent(FaultDetectionListener.EVENT_INTERRUPTED,
            localAddress);
      }

      // Get results
      synchronized (addresses)
      {
        for (int i = 0; i < addresses.size(); i++)
          getResult((IpAddress) addresses.get(i));
      }
    }

    // Quitting
    notifyListenerOfEvent(FaultDetectionListener.EVENT_QUIT, localAddress);
  }
}