/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.gms;

import org.objectweb.tribe.channel.ReceiveBuffer;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.EmptyBufferException;
import org.objectweb.tribe.gms.protocol.GroupMembershipMessage;
import org.objectweb.tribe.messages.ChannelMessage;

/**
 * This class defines a GroupMembershipListenerThread.
 * <p>
 * This threads listen for group membership messages and call the appropriate
 * callbacks.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class GroupMembershipListenerThread extends Thread
{
  private GroupMembershipService listener;
  private ReceiveBuffer          receiveBuffer;   // where GMS messages are
  // delivered
  private boolean                isKilled = false;
  private Trace                  logger;

  /**
   * Creates a new <code>GroupMembershipListenerThread</code> object
   */
  public GroupMembershipListenerThread(GroupMembershipService gms)
  {
    super("GroupMembershipListenerThread");
    this.listener = gms;
    logger = GroupMembershipService.getLogger();
  }

  /**
   * Terminates the GroupMembershipListenerThread.
   */
  public void kill()
  {
    isKilled = true;
    // Wake up thread with a null message
    receiveBuffer.addMessage(null);
  }

  /**
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    if (logger.isDebugEnabled())
      logger.debug("GroupMembershipListenerThread started");
    receiveBuffer = listener.getReceiveBuffer();
    while (!isKilled)
    {
      try
      {
        ChannelMessage msg = receiveBuffer.getMessage();
        if (msg instanceof GroupMembershipMessage)
        {
          if (logger.isDebugEnabled())
            logger.debug("Delivering GMS message: " + msg);
          GroupMembershipMessage gmsMsg = (GroupMembershipMessage) msg;
          gmsMsg.deliver(listener);
        }
        else
          logger
              .debug("Unexpected message type in GroupMembershipListenerThread:"
                  + msg);
      }
      catch (EmptyBufferException e)
      {
        logger.error("Error while getting message for GMS service", e);
      }
    }
  }
}