/**
 * Tribe: Group communication library.
 * Copyright (C) 2004-2005 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.gms;

import java.util.ArrayList;
import java.util.HashMap;

import org.objectweb.tribe.channel.AbstractChannelPool;
import org.objectweb.tribe.channel.AbstractServerChannel;
import org.objectweb.tribe.channel.ReceiveBuffer;
import org.objectweb.tribe.channel.ReliableGroupChannelWithGms;
import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.gms.discovery.DiscoveryListener;
import org.objectweb.tribe.gms.discovery.DiscoveryService;
import org.objectweb.tribe.gms.discovery.UdpDiscoveryService;
import org.objectweb.tribe.gms.protocol.GroupCompositionMessage;
import org.objectweb.tribe.gms.protocol.QuitMessage;
import org.objectweb.tribe.messages.PingMessage;

/**
 * This class defines a GroupMembershipService
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class GroupMembershipService
    implements
      DiscoveryListener,
      GroupMembershipListener
{
  /** Destination key for GMS messages */
  public static String                  GMS_KEY = "tribe.gms";

  // HashMap: GroupIdentifier -> Group
  protected HashMap                     groupMemberships;
  private GroupMembershipListenerThread membershipThread;
  private IpAddress                     replyAddress;
  private AbstractServerChannel         serverChannel;
  private DiscoveryService              discovery;
  protected ArrayList                   listeners;
  private AbstractChannelPool           channelPool;
  private ReceiveBuffer                 receiveBuffer;
  protected static Trace                logger  = Trace
                                                    .getLogger("org.objectweb.tribe.gms");

  /**
   * Creates a new <code>GroupMembershipService</code> object
   * 
   * @param replyAddress address to send replies to
   * @param channelPool channel pool for group communications
   * @param discovery the discovery service to use
   * @throws ChannelException if no channel can be bound to the reply address
   */
  public GroupMembershipService(IpAddress replyAddress,
      AbstractChannelPool channelPool, DiscoveryService discovery)
      throws ChannelException
  {
    groupMemberships = new HashMap();
    listeners = new ArrayList();
    this.channelPool = channelPool;
    this.discovery = discovery;
    if (channelPool != null)
    {
      receiveBuffer = new ReceiveBuffer(GroupMembershipService.GMS_KEY);
      membershipThread = new GroupMembershipListenerThread(this);
      membershipThread.start();
      channelPool.registerReceiveBuffer(receiveBuffer);
      serverChannel = channelPool.getServerChannel(replyAddress);
      try
      {
        // Get the real bind address (especially if port was set to 0)
        this.replyAddress = (IpAddress) serverChannel.getBindAddress();
        if (this.replyAddress.getAddress().getHostAddress().equals("0.0.0.0"))
          this.replyAddress.setAddress(replyAddress.getAddress());
      }
      catch (NotConnectedException e)
      {
        throw new ChannelException("Unable to bind server channel.");
      }

      discovery.registerDiscoveryListener(this);
      ((UdpDiscoveryService) discovery).setReplyAddress(this.replyAddress);
    }
  }

  /**
   * Returns the receiveBuffer value.
   * 
   * @return Returns the receiveBuffer.
   */
  protected ReceiveBuffer getReceiveBuffer()
  {
    return receiveBuffer;
  }

  /**
   * Returns the logger value.
   * 
   * @return Returns the logger.
   */
  protected static Trace getLogger()
  {
    return logger;
  }

  /**
   * Joins the group that has the given group identifier.
   * 
   * @param channel client channel to join the group
   * @param gid the identifier of the group to join
   * @return the Member corresponding to the newly joining member
   * @throws AlreadyMemberException if we are already member of the group
   * @throws NotConnectedException if the channel is not connected
   * @throws ChannelException if an error is reported by the channel
   */
  public Member join(ReliableGroupChannelWithGms channel, GroupIdentifier gid)
      throws AlreadyMemberException, NotConnectedException, ChannelException
  {
    Member me = new Member(replyAddress, channel.toString());
    if (logger.isDebugEnabled())
      logger.debug("Member " + me.getUid() + " joins group " + gid);
    Group g;
    synchronized (groupMemberships)
    {
      g = getGroup(gid);
      if (g == null)
      { // Group does not exist, create it
        g = new Group(gid);
        groupMemberships.put(gid, g);
      }
      else if (g.hasMember(me))
        throw new AlreadyMemberException();

      // Add ourselves to the group
      g.addMember(me);
    }

    // Ask for the group composition
    if (discovery != null)
      discovery.sendGroupDiscovery(gid);

    // Notify listeners
    synchronized (listeners)
    {
      int size = listeners.size();
      for (int i = 0; i < size; i++)
        ((GroupMembershipListener) listeners.get(i)).joinMember(me, gid);
    }

    return me;
  }

  /**
   * Leaves the group that has the given group identifier.
   * 
   * @param channel client channel to leave the group
   * @param gid the identifier of the group to quit
   * @throws ChannelException if an error occurs
   * @throws NotConnectedException if m does not belong to this group
   */
  public void quit(ReliableGroupChannelWithGms channel, GroupIdentifier gid)
      throws ChannelException, NotConnectedException
  {
    Group g = getGroup(gid);
    if (g == null)
      throw new NotConnectedException(
          "Trying to leave a group we do not belong to.");
    Member me = new Member(replyAddress, channel.toString());
    if (logger.isDebugEnabled())
      logger.debug("Member " + me.getUid() + " quits group " + gid);
    if (channelPool != null)
    {
      ArrayList failed = channelPool.send(new QuitMessage(gid, me), g
          .getMembers());
      if (failed != null)
        throw new ChannelException(failed.size()
            + " member did not receive the quit message.");
    }
  }

  /**
   * Get the list of members in a group.
   * 
   * @param gid the group identifier
   * @return the <code>Group</code> corresponding to this gid or null if no
   *         such group is registered in the GMS.
   */
  public Group getGroup(GroupIdentifier gid)
  {
    synchronized (groupMemberships)
    {
      return (Group) groupMemberships.get(gid);
    }
  }

  /**
   * Register a new GroupMembershipListener.
   * 
   * @param listener the listener to add
   */
  public void registerGroupMembershipListener(GroupMembershipListener listener)
  {
    synchronized (listeners)
    {
      listeners.add(listener);
    }
  }

  /**
   * Unregister a GroupMembershipListener.
   * 
   * @param listener the listener to remove
   * @return true if the listener was registered, false otherwise
   */
  public boolean unregisterGroupMembershipListener(
      GroupMembershipListener listener)
  {
    synchronized (listeners)
    {
      return listeners.remove(listener);
    }
  }

  //
  // Discovery Service related functions
  //

  /**
   * @see org.objectweb.tribe.gms.discovery.DiscoveryListener#discoveryRequest(org.objectweb.tribe.common.GroupIdentifier,
   *      org.objectweb.tribe.common.Address)
   */
  public void discoveryRequest(GroupIdentifier gid, Address sender)
  {
    Group g = getGroup(gid);
    if (g != null)
    {
      if (logger.isDebugEnabled())
        logger.debug("Sending GroupCompositionMessage for group "
            + g.getGroupIdentifier() + " to " + sender);

      try
      {
        channelPool.getChannel(sender).send(
            new GroupCompositionMessage(g, replyAddress));
      }
      catch (ChannelException e)
      {
        logger.error("Failed to send GroupCompositionMessage for group " + gid
            + " to " + sender, e);
      }
      catch (NotConnectedException e)
      {
        logger.error("Unable to read " + sender + " to send group " + gid
            + " composition.");
      }
    }
    else if (logger.isDebugEnabled())
      logger.debug("No information to send for group " + gid);

  }

  //
  //  Group Membership Listener related functions
  //

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#groupComposition(Group,
   *      Address)
   */
  public void groupComposition(Group g, Address sender)
  {
    // Discard local messages
    //  if (replyAddress.equals(sender))
    //  return;

    GroupIdentifier gid = g.getGroupIdentifier();
    synchronized (groupMemberships)
    {
      Group localGroup = getGroup(gid);
      if (localGroup == null)
      {
        if (logger.isDebugEnabled())
          logger.debug("New group " + gid + " composition");
        groupMemberships.put(gid, g);
      }
      else
      {
        if (logger.isDebugEnabled())
          logger
              .debug("Merging local group composition with the one received from "
                  + sender);
        if (localGroup.merge(g))
        { // Group composition has changed, broadcast the new group composition
          // to everybody in the group.

          // First check that all members in the new group are reachable
          if (logger.isDebugEnabled())
            logger.debug("Checking new group composition:"
                + localGroup.getStringMembers());
          ArrayList failed = channelPool.send(new PingMessage(), localGroup
              .getMembers());
          if (failed != null)
          { // Removed all failed members from group
            ArrayList remainingMembers = localGroup.getMembers();
            remainingMembers.removeAll(failed);
            // Annonced that these members failed
            for (int i = 0; i < failed.size(); i++)
            {
              channelPool.send(new QuitMessage(gid, (Member) failed.get(i)),
                  remainingMembers);
            }
          }

          // Now broadcast the real group composition with only alive members
          if (logger.isDebugEnabled())
            logger.debug("Broacasting new group composition:"
                + localGroup.getStringMembers());
          failed = channelPool.send(new GroupCompositionMessage(localGroup,
              replyAddress), localGroup.getMembers());
          if (failed != null)
          { // Removed all failed members from group
            ArrayList remainingMembers = localGroup.getMembers();
            remainingMembers.removeAll(failed);
            // Annonced that these members failed
            for (int i = 0; i < failed.size(); i++)
            {
              channelPool.send(new QuitMessage(gid, (Member) failed.get(i)),
                  remainingMembers);
            }
          }
        }
        else if (logger.isDebugEnabled())
        {
          logger.debug("Group composition has not changed.");
        }
      }
    }

    // Notify listeners
    synchronized (listeners)
    {
      int size = listeners.size();
      for (int i = 0; i < size; i++)
        ((GroupMembershipListener) listeners.get(i))
            .groupComposition(g, sender);
    }
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#joinMember(org.objectweb.tribe.common.Member,
   *      org.objectweb.tribe.common.GroupIdentifier)
   */
  public void joinMember(Member m, GroupIdentifier gid)
  {
    synchronized (groupMemberships)
    {
      Group localGroup = getGroup(gid);
      if (localGroup == null)
      {
        if (logger.isDebugEnabled())
          logger.debug("Ignoring member " + m + " joining group " + gid);
      }
      else
      {
        if (logger.isDebugEnabled())
          logger.debug("Member " + m + " joins group " + gid);
        localGroup.addMember(m);
      }
    }

    // Notify listeners
    synchronized (listeners)
    {
      int size = listeners.size();
      for (int i = 0; i < size; i++)
        ((GroupMembershipListener) listeners.get(i)).joinMember(m, gid);
    }
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#quitMember(org.objectweb.tribe.common.Member,
   *      org.objectweb.tribe.common.GroupIdentifier)
   */
  public void quitMember(Member m, GroupIdentifier gid)
  {
    synchronized (groupMemberships)
    {
      Group localGroup = getGroup(gid);
      if (localGroup == null)
      {
        if (logger.isDebugEnabled())
          logger.debug("Ignoring member " + m + " leaving group " + gid);
      }
      else
      {
        if (logger.isDebugEnabled())
          logger.debug("Member " + m + " leaves group " + gid);
        localGroup.removeMember(m);
      }
    }

    // Notify listeners
    synchronized (listeners)
    {
      int size = listeners.size();
      for (int i = 0; i < size; i++)
        ((GroupMembershipListener) listeners.get(i)).quitMember(m, gid);
    }
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#failedMember(org.objectweb.tribe.common.Member,
   *      org.objectweb.tribe.common.GroupIdentifier,
   *      org.objectweb.tribe.common.Member)
   */
  public void failedMember(Member failed, GroupIdentifier gid, Member sender)
  {
    synchronized (groupMemberships)
    {
      Group localGroup = getGroup(gid);
      if (localGroup == null)
      {
        if (logger.isDebugEnabled())
          logger.debug("Ignoring member " + failed + " failed in group " + gid);
      }
      else
      {
        if (logger.isInfoEnabled())
          logger.info("Member " + failed + " failed in group " + gid);
        localGroup.removeMember(failed);
      }
    }

    // Notify listeners
    synchronized (listeners)
    {
      int size = listeners.size();
      for (int i = 0; i < size; i++)
        ((GroupMembershipListener) listeners.get(i)).failedMember(failed, gid,
            sender);
    }
  }

  /**
   * Stop this GMS service and its associated thread (note that the discovery
   * service must be stopped separately).
   */
  public void stop()
  {
    groupMemberships.clear();
    if (channelPool != null)
    {
      membershipThread.kill();
      channelPool.removeServerChannelFromPool(serverChannel);
      try
      {
        membershipThread.join(1000);
      }
      catch (InterruptedException ignore)
      {
      }
    }
  }

}