/**
 * C-JDBC: Clustered JDBC.
 * Copyright (C) 2002-2005 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: c-jdbc@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.gms;

import java.net.URL;
import java.util.Vector;

import org.jgroups.ChannelClosedException;
import org.jgroups.JChannel;
import org.jgroups.MembershipListener;
import org.jgroups.View;
import org.objectweb.tribe.channel.ReliableGroupChannelWithGms;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;

/**
 * This class defines a JGroupsMembershipService
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class JGroupsMembershipService extends GroupMembershipService
    implements
      MembershipListener
{
  // The underlying JGroups channel with GMS
  private JChannel        jgroupsChannel;
  private GroupIdentifier currentGroup;

  /**
   * Creates a new <code>JGroupsMembershipService</code> object
   * 
   * @param jgroupsConfigFile JGroups configuration file
   */
  public JGroupsMembershipService(URL jgroupsConfigFile)
      throws ChannelException
  {
    super(null, null, null);
    try
    {
      jgroupsChannel = new JChannel(jgroupsConfigFile);
    }
    catch (org.jgroups.ChannelException e)
    {
      throw new ChannelException(e);
    }
  }

  /**
   * Gets underlying JGroups channel.
   * 
   * @return the JGroups channel
   */
  public JChannel getJGroupsChannel()
  {
    return jgroupsChannel;
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipService#join(org.objectweb.tribe.channel.ReliableGroupChannelWithGms,
   *      org.objectweb.tribe.common.GroupIdentifier)
   */
  public Member join(ReliableGroupChannelWithGms channel, GroupIdentifier gid)
      throws AlreadyMemberException, NotConnectedException, ChannelException
  {
    Member me;
    Group g;
    synchronized (groupMemberships)
    {
      try
      {
        jgroupsChannel.connect(gid.getGroupName());
        currentGroup = gid;
      }
      catch (ChannelClosedException e)
      {
        throw new NotConnectedException(e);
      }
      catch (org.jgroups.ChannelException e)
      {
        throw new ChannelException(e);
      }

      g = getGroup(currentGroup);
      me = memberFromJGroupsAddress((org.jgroups.stack.IpAddress) jgroupsChannel
          .getLocalAddress());
      if (g == null)
      { // Group does not exist, create it
        g = new Group(currentGroup);
        groupMemberships.put(gid, g);
      }
      else
        throw new AlreadyMemberException();

      // Add ourselves to the group
      g.addMember(me);
    }

    // Notify listeners
    synchronized (listeners)
    {
      int size = listeners.size();
      for (int i = 0; i < size; i++)
        ((GroupMembershipListener) listeners.get(i)).joinMember(me, gid);
    }

    return me;
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipService#quit(org.objectweb.tribe.channel.ReliableGroupChannelWithGms,
   *      org.objectweb.tribe.common.GroupIdentifier)
   */
  public void quit(ReliableGroupChannelWithGms channel, GroupIdentifier gid)
      throws ChannelException, NotConnectedException
  {
    super.quit(channel, gid);

    jgroupsChannel.disconnect();
  }

  //
  // JGroups MembershipListener interface implementation
  //

  /**
   * Returns a tribe Member object from a JGroups address (supposed to be an
   * IpAddress).
   * 
   * @param jgroupsAddress the JGroups address
   * @return a tribe Member corresponding to this JGroups address
   */
  public static Member memberFromJGroupsAddress(
      org.jgroups.stack.IpAddress jgroupsAddress)
  {
    return new Member(new IpAddress(jgroupsAddress.getIpAddress(),
        jgroupsAddress.getPort()), jgroupsAddress.getIpAddress().toString()
        + ":" + jgroupsAddress.getPort());
  }

  /**
   * @see org.jgroups.MembershipListener#viewAccepted(org.jgroups.View)
   */
  public void viewAccepted(View newView)
  {
    if (logger.isDebugEnabled())
      logger.debug("JGroups reported new view: " + newView);
    Vector members = newView.getMembers();
    int size = members.size();
    synchronized (groupMemberships)
    {
      Group g = getGroup(currentGroup);
      for (int i = 0; i < size; i++)
      {
        Member m = memberFromJGroupsAddress((org.jgroups.stack.IpAddress) members
            .get(i));
        if (!g.hasMember(m)) // Add new member
          joinMember(m, currentGroup);
      }
    }
  }

  /**
   * @see org.jgroups.MembershipListener#suspect(org.jgroups.Address)
   */
  public void suspect(org.jgroups.Address suspectedMbr)
  {
    if (logger.isDebugEnabled())
      logger.debug("JGroups reports suspected member:" + suspectedMbr);
    org.jgroups.stack.IpAddress jgroupsAddress = (org.jgroups.stack.IpAddress) suspectedMbr;
    quitMember(memberFromJGroupsAddress(jgroupsAddress), currentGroup);
  }

  /**
   * @see org.jgroups.MembershipListener#block()
   */
  public void block()
  {
    if (logger.isDebugEnabled())
      logger.debug("JGroups reported block()");
  }

}