/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.gms.discovery;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.MulticastSocket;
import java.net.SocketException;
import java.util.ArrayList;

import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.log.Trace;
import org.objectweb.tribe.gms.protocol.GroupDiscoveryMessage;
import org.objectweb.tribe.messages.DatagramMessage;

/**
 * This class defines a UdpDiscoveryService.
 * <p>
 * <code>GroupDiscoveryMessage</code> are sent on the given multicast address.
 * All registered listeners are called when a GroupDiscoveryMessage is received.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class UdpDiscoveryService extends Thread implements DiscoveryService
{
  private static final int RECEIVE_BUFFER_SIZE = 1024;

  private MulticastSocket  multicastSocket;
  private IpAddress        multicastAddress;
  private IpAddress        replyAddress;
  private ArrayList        listeners;
  private boolean          isKilled            = false;

  private Trace            logger              = Trace
                                                   .getLogger("org.objectweb.tribe.discovery");

  /**
   * Creates a new <code>UdpDiscoveryService</code> object
   * 
   * @param multicastAddress multicast address to use for discovery messages.
   * @param inChannel the address on which responses must be sent back.
   */
  public UdpDiscoveryService(IpAddress multicastAddress, IpAddress replyAddress)
  {
    super("UdpDiscoveryService");
    if (multicastAddress == null)
      throw new IllegalArgumentException(
          "Null addresses in UdpDiscoveryService constructor");

    listeners = new ArrayList();
    this.replyAddress = replyAddress;
    try
    {
      // Create the multicast socket
      logger.info("Creating UdpDiscoveryService on " + multicastAddress);
      this.multicastAddress = multicastAddress;
      multicastSocket = new MulticastSocket(multicastAddress.getPort());
      multicastSocket.joinGroup(multicastAddress.getAddress());
      start();
    }
    catch (SocketException e)
    {
      logger.error("Failed to create multicast socket", e);
    }
    catch (IOException e)
    {
      logger.error("Failed to create multicast socket", e);
    }
  }

  /**
   * @see org.objectweb.tribe.gms.discovery.DiscoveryService#sendGroupDiscovery(org.objectweb.tribe.common.GroupIdentifier)
   */
  public void sendGroupDiscovery(GroupIdentifier gid)
  {
    try
    {
      if (logger.isDebugEnabled())
        logger.debug("Sending GroupDiscoveryMessage for group " + gid
            + " reply to " + replyAddress);
      DatagramPacket joinPacket = new GroupDiscoveryMessage(replyAddress,
          multicastAddress, gid).getDatagramPacket();
      multicastSocket.send(joinPacket);
    }
    catch (IOException e)
    {
      if (logger.isDebugEnabled())
        logger.debug("Failed to send group discovery message for group " + gid,
            e);
      else if (logger.isInfoEnabled())
        logger.info("Failed to send group discovery message for group " + gid
            + " (" + e + ")");
    }
  }

  /**
   * @see org.objectweb.tribe.gms.discovery.DiscoveryService#registerDiscoveryListener(org.objectweb.tribe.gms.discovery.DiscoveryListener)
   */
  public void registerDiscoveryListener(DiscoveryListener listener)
  {
    synchronized (listeners)
    {
      listeners.add(listener);
    }
  }

  /**
   * @see org.objectweb.tribe.gms.discovery.DiscoveryService#unregisterDiscoveryListener(org.objectweb.tribe.gms.discovery.DiscoveryListener)
   */
  public boolean unregisterDiscoveryListener(DiscoveryListener listener)
  {
    synchronized (listeners)
    {
      return listeners.remove(listener);
    }
  }

  /**
   * The DiscoveryService threads looks for JOIN messages and notify appropriate
   * listeners.
   */
  public void run()
  {
    DatagramPacket datagram = new DatagramPacket(new byte[RECEIVE_BUFFER_SIZE],
        RECEIVE_BUFFER_SIZE);

    while (!isKilled)
    {
      try
      {
        multicastSocket.receive(datagram);
        Object received = DatagramMessage.getObjectFromDatagram(datagram);
        if (received != null)
        {
          if (received instanceof GroupDiscoveryMessage)
          {
            GroupDiscoveryMessage msg = (GroupDiscoveryMessage) received;
            if (logger.isDebugEnabled())
              logger.debug("Received GroupDiscoveryMessage for group "
                  + msg.getGroupIdentifier() + " from "
                  + msg.getSourceAddress());
            // Notify listeners
            synchronized (listeners)
            {
              int size = listeners.size();
              for (int i = 0; i < size; i++)
              {
                DiscoveryListener listener = (DiscoveryListener) listeners
                    .get(i);
                listener.discoveryRequest(msg.getGroupIdentifier(), msg
                    .getSourceAddress());
              }
            }
          }
          else
            logger.info("UdpDiscoveryService received an unexpected message: "
                + received);
        }
      }
      catch (IOException e)
      {
        logger.error("Error while receiving message", e);
      }
    }
  }

  /**
   * Returns the replyAddress value.
   * 
   * @return Returns the replyAddress.
   */
  public IpAddress getReplyAddress()
  {
    return replyAddress;
  }

  /**
   * Sets the replyAddress value.
   * 
   * @param replyAddress The replyAddress to set.
   */
  public void setReplyAddress(IpAddress replyAddress)
  {
    this.replyAddress = replyAddress;
  }

  /**
   * Terminate the UdpDiscoveryService.
   */
  public void kill()
  {
    isKilled = true;
    multicastSocket.close();
    this.interrupt();
    try
    {
      this.join(1000);
    }
    catch (InterruptedException e)
    {
    }
  }
}