/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.messages;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import org.objectweb.tribe.channel.ReceiveBuffer;
import org.objectweb.tribe.exceptions.NoReceiverException;

/**
 * This class defines a ChannelMessage.
 * <p>
 * A ChannelMessage is made of chunks, the last chunk of the message containing
 * the destination key to be used for message routing at reception.
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public abstract class ChannelMessage implements Serializable
{
  private Object           key;                  // Message destination key
  private ArrayList        chunks;
  private transient byte[] messageInBytes = null;

  /**
   * Creates a new <code>ChannelMessage</code> object
   * 
   * @param destinationKey receive buffer key for message delivery
   */
  public ChannelMessage(Object destinationKey)
  {
    chunks = new ArrayList();
    this.key = destinationKey;
  }

  /**
   * Creates a new <code>ChannelMessage</code> object
   * 
   * @param destinationKey receive buffer key for message delivery
   * @param numberOfChunks number of chunks in the message if known in advance
   */
  public ChannelMessage(Object destinationKey, int numberOfChunks)
  {
    chunks = new ArrayList(numberOfChunks);
    this.key = destinationKey;
  }

  /**
   * Adds a chunk to this message
   * 
   * @param chunk the message chunk to add
   */
  public synchronized void addChunk(Serializable chunk)
  {
    messageInBytes = null;
    chunks.add(chunk);
  }

  /**
   * Get the whole list of chunks. Warning this is not a copy but the real
   * message content.
   * 
   * @return an <code>ArrayList</code> of chunks (Objects).
   */
  public ArrayList getChunks()
  {
    return chunks;
  }

  /**
   * Remove the last chunk.
   * 
   * @return the last chunk of the chunk list or null if no more chunks are
   *         available.
   */
  public synchronized Object removeChunk()
  {
    messageInBytes = null;
    int size = chunks.size();
    if (size == 0)
      return null;
    else
      return chunks.remove(size - 1);
  }

  /**
   * Defines the destination key used for message delivery at the receiving
   * host. A <code>ReceiveBuffer</code> must has been registered with this
   * destination key at the receiver end for the message to be delivered.
   * 
   * @return the destination key
   */
  public Object getDestinationKey()
  {
    return key;
  }

  /**
   * Delivers the message to all receive buffers that are registered to the
   * message destination key.
   * 
   * @param receiveBuffers list of destination receive buffers
   * @throws NoReceiverException if no receive buffer has been registered for
   *           the destination key found in the message
   */
  public void deliverMessage(HashMap receiveBuffers) throws NoReceiverException
  {
    ArrayList buffers;
    synchronized (receiveBuffers)
    {
      // Get buffers for this gid
      buffers = (ArrayList) receiveBuffers.get(key);
    }
    if (buffers == null)
      throw new NoReceiverException("No receive buffer to deliver message for "
          + key);

    // Deliver the message to all buffers
    // TODO: Use a reader/writer lock instead for better scalability
    synchronized (buffers)
    {
      int size = buffers.size();
      for (int i = 0; i < size; i++)
      {
        ReceiveBuffer buf = (ReceiveBuffer) buffers.get(i);
        buf.addMessage(this);
      }
    }
  }

  /**
   * Get a byte array representation of the message. This representation is
   * computed only once (which means that the message is serialized only once)
   * but calls to add/remove chunks will force a recompute of the message
   * representation.
   * 
   * @return this message as a byte array
   */
  public synchronized byte[] getByteArray()
  {
    if (messageInBytes == null)
    {
      try
      {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        new ObjectOutputStream(baos).writeObject(this);
        return baos.toByteArray();
      }
      catch (IOException e)
      {
        messageInBytes = null;
        e.printStackTrace();
      }
    }
    return messageInBytes;
  }

}