/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.messages;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.DatagramPacket;

import org.objectweb.tribe.common.IpAddress;

/**
 * This class defines a DatagramMessage
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class DatagramMessage implements Serializable
{
  private IpAddress sourceAddress;
  private IpAddress destinationAddress;
  private byte[]    content = null;

  /**
   * Creates a new <code>DatagramMessage</code> object
   * 
   * @param sourceAddress source address
   * @param destinationAddress destination address
   */
  public DatagramMessage(IpAddress sourceAddress, IpAddress destinationAddress)
  {
    this.sourceAddress = sourceAddress;
    this.destinationAddress = destinationAddress;
  }

  /**
   * Returns the destinationAddress value.
   * 
   * @return Returns the destinationAddress.
   */
  public IpAddress getDestinationAddress()
  {
    return destinationAddress;
  }

  /**
   * Returns the sourceAddress value.
   * 
   * @return Returns the sourceAddress.
   */
  public IpAddress getSourceAddress()
  {
    return sourceAddress;
  }

  /**
   * Returns the content value.
   * 
   * @return Returns the content.
   */
  public byte[] getContent()
  {
    return content;
  }

  /**
   * Sets the content value.
   * 
   * @param content The content to set.
   */
  public void setContent(byte[] content)
  {
    this.content = content;
  }

  /**
   * Returns the message content size
   * 
   * @return message content size
   */
  public int getContentSize()
  {
    if (content == null)
      return 0;
    else
      return content.length;
  }

  /**
   * Return the current message as a DatagramPacket to send.
   * 
   * @return a DatagramPacket corresponding to the current message.
   */
  public DatagramPacket getDatagramPacket()
  {
    if (content == null)
      content = objectToBytes(this);
    return new DatagramPacket(content, content.length, getDestinationAddress()
        .getAddress(), getDestinationAddress().getPort());
  }

  /**
   * Return the content of this datagram as an object.
   * 
   * @param datagram the datagram to convert.
   * @return the Object contained in the datagram or null if the conversion
   *         fails.
   */
  public static Object getObjectFromDatagram(DatagramPacket datagram)
  {
    try
    {
      ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(
          datagram.getData()));
      return in.readObject();
    }
    catch (IOException e)
    {
      return null;
    }
    catch (ClassNotFoundException e)
    {
      return null;
    }
  }

  /**
   * Converts the given object to an array of bytes.
   * 
   * @param obj the object to convert
   * @return an array of bytes
   */
  public static byte[] objectToBytes(Serializable obj)
  {
    if (obj == null)
      return null;
    try
    {
      ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
      new ObjectOutputStream(byteStream).writeObject(obj);
      return byteStream.toByteArray();
    }
    catch (IOException ex)
    {
      throw new IllegalArgumentException(ex.toString());
    }
  }

}