/**
 * Tribe: Group communication library.
 * Copyright (C) 2002-2005 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.adapters;

import java.io.Serializable;
import java.net.InetAddress;
import java.util.ArrayList;

import org.objectweb.tribe.channel.ReliableGroupChannelWithGms;
import org.objectweb.tribe.channel.tcp.TcpChannelPool;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.exceptions.ChannelException;
import org.objectweb.tribe.exceptions.NotConnectedException;
import org.objectweb.tribe.gms.GroupMembershipService;
import org.objectweb.tribe.gms.discovery.UdpDiscoveryService;
import org.objectweb.tribe.messages.MessageListener;

/**
 * This class defines a MulticastRequestAdapterTest
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class MulticastRequestAdapterTest
    implements
      MessageListener,
      MulticastRequestListener
{

  private ReliableGroupChannelWithGms channel1;
  private ReliableGroupChannelWithGms channel2;
  private UdpDiscoveryService         discovery;
  private GroupMembershipService      gms;
  private final GroupIdentifier       gid1 = new GroupIdentifier("group1");
  private final String                msg1 = "Hello from 1";
  private final String                msg2 = "Hello from 2";
  private MulticastRequestAdapter     adapter1;
  private MulticastRequestAdapter     adapter2;

  /**
   * Creates a new <code>MulticastRequestAdapterTest</code> object
   * 
   * @throws Exception if an error occurs
   */
  public MulticastRequestAdapterTest() throws Exception
  {
    initializeChannels();
  }

  private void initializeChannels() throws Exception
  {
    final InetAddress MULTICAST_ADDRESS = InetAddress.getByName("224.7.65.23");
    final int MULTICAST_PORT = 2288;
    final IpAddress MULTICAST_IP = new IpAddress(MULTICAST_ADDRESS,
        MULTICAST_PORT);
    final InetAddress REPLY_ADDRESS = InetAddress.getLocalHost();
    final int REPLY_PORT = 0; // Let the system choose a port
    final IpAddress REPLY_IP = new IpAddress(REPLY_ADDRESS, REPLY_PORT);

    discovery = new UdpDiscoveryService(MULTICAST_IP, REPLY_IP);
    gms = new GroupMembershipService(REPLY_IP, TcpChannelPool.getChannelPool(),
        discovery);
    channel1 = new ReliableGroupChannelWithGms(gms);
    channel2 = new ReliableGroupChannelWithGms(gms);

    // Join

    try
    {
      channel1.join(gid1);
    }
    catch (AlreadyMemberException e)
    {
      e.printStackTrace();
      throw e;
    }
    try
    {
      channel2.join(gid1);
    }
    catch (AlreadyMemberException e)
    {
      e.printStackTrace();
      throw e;
    }

    adapter1 = new MulticastRequestAdapter(channel1, this, this);
    adapter2 = new MulticastRequestAdapter(channel2, this, this);

  }

  /**
   * Test the group join/send/receive/quit primitives
   */
  public void sendMessages()
  {
    try
    {
      ArrayList members = channel1.getCurrentGroup().getMembers();
      // Send msg1 on channel1
      System.out.println("Sending '" + msg1 + "' to " + members.size()
          + " members.");
      MulticastResponse replies = adapter1.multicastMessage(members, msg1,
          MulticastRequestAdapter.WAIT_ALL, 0);

      // Display replies
      for (int i = 0; i < members.size(); i++)
      {
        Member m = (Member) members.get(i);
        Object answer = replies.getResult(m);
        if (answer == null)
          System.out.println("Member " + m + " failed.");
        else
          System.out.println("Member " + m + " replied: '" + answer + "'");
      }

      // Send msg2 on channel2
      System.out.println("Sending '" + msg2 + "' to " + members.size()
          + " members.");
      adapter2.multicastMessage(members, msg2,
          MulticastRequestAdapter.WAIT_ALL, 0);

      // Display replies
      for (int i = 0; i < members.size(); i++)
      {
        Member m = (Member) members.get(i);
        Object answer = replies.getResult(m);
        if (answer == null)
          System.out.println("Member " + m + " failed.");
        else
          System.out.println("Member " + m + " replied: '" + answer + "'");
      }
    }
    catch (Exception e1)
    {
      e1.printStackTrace();
    }
  }

  /**
   * Quit the group and closes the channels.
   */
  public void close()
  {
    // Stop the adapters
    System.out.println("Stopping adapter1");
    adapter1.stop();
    System.out.println("Stopping adapter2");
    adapter2.stop();

    // Stop GMS and discovery
    System.out.println("Stopping GMS");
    gms.stop();
    System.out.println("Stopping discovery service");
    discovery.kill();

    // Close the channels
    try
    {
      System.out.println("Closing channel1");
      channel1.close();
    }
    catch (ChannelException e)
    {
      e.printStackTrace();
    }
    catch (NotConnectedException e)
    {
      e.printStackTrace();
    }
    try
    {
      System.out.println("Closing channel2");
      channel2.close();
    }
    catch (ChannelException e1)
    {
      e1.printStackTrace();
    }
    catch (NotConnectedException e1)
    {
      e1.printStackTrace();
    }

  }

  //
  // Message listener interface
  //

  /**
   * @see org.objectweb.tribe.messages.MessageListener#receive(java.io.Serializable)
   */
  public void receive(Serializable msg)
  {
    System.out.println("Received msg: " + msg);
  }

  //
  // MulticastRequestListener interface
  //

  /**
   * Main method
   * 
   * @param args command line arguments
   */
  public static void main(String[] args)
  {
    try
    {
      MulticastRequestAdapterTest adapter = new MulticastRequestAdapterTest();
      adapter.sendMessages();
      adapter.sendMessages();
      adapter.sendMessages();
      adapter.close();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  /**
   * @see org.objectweb.tribe.adapters.MulticastRequestListener#handleMessageSingleThreaded(java.io.Serializable,
   *      org.objectweb.tribe.common.Member)
   */
  public Object handleMessageSingleThreaded(Serializable msg, Member sender)
  {
    System.out.println("Received single threaded '" + msg + "' from " + sender);
    return "singleDone";
  }

  /**
   * @see org.objectweb.tribe.adapters.MulticastRequestListener#handleMessageMultiThreaded(java.io.Serializable,
   *      org.objectweb.tribe.common.Member, java.lang.Object)
   */
  public Serializable handleMessageMultiThreaded(Serializable msg,
      Member sender, Object handleMessageSingleThreadedResult)
  {
    System.out.println("Received multithreaded '" + msg + "' from " + sender
        + " (" + handleMessageSingleThreadedResult + ")");
    return "singleMulti";
  }

}