/**
 * Tribe: Group communication library.
 * Copyright (C) 2002-2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package org.objectweb.tribe.channel;

import java.io.Serializable;
import java.net.InetAddress;
import java.util.ArrayList;

import org.objectweb.tribe.adapters.PullPushAdapter;
import org.objectweb.tribe.channel.tcp.TcpChannelPool;
import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.exceptions.AlreadyMemberException;
import org.objectweb.tribe.gms.GroupMembershipListener;
import org.objectweb.tribe.gms.GroupMembershipService;
import org.objectweb.tribe.gms.discovery.UdpDiscoveryService;
import org.objectweb.tribe.messages.ByteMessage;
import org.objectweb.tribe.messages.MessageListener;

/**
 * This class defines a ChannelPerformaceTest
 * 
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class ChannelPerformaceTest
    implements
      MessageListener,
      GroupMembershipListener
{
  private ReliableGroupChannelWithGms channel;
  private UdpDiscoveryService         discovery;
  private GroupMembershipService      gms;
  private final GroupIdentifier       testGid = new GroupIdentifier("groupTest");
  private PullPushAdapter             adapter;
  private int                         received;
  private int                         msgSize;
  private int                         nbOfMessages;
  private int                         nbOfSenders;
  private int                         nbOfMembers;
  private int                         done;
  private long                        start;
  private boolean                     sending = false;

  /**
   * Creates a new <code>ChannelPerformaceTest</code> object
   * 
   * @param members
   * @param senders
   * @param messages
   * @param size
   */
  public ChannelPerformaceTest(int members, int senders, int messages, int size)
  {
    if (members == 0)
    {
      System.out.println("Members must be >0.");
      System.exit(1);
    }
    if (senders == 0)
    {
      System.out.println("Senders must be >0.");
      System.exit(1);
    }
    if (senders > members)
    {
      System.out
          .println("Senders must be lower or equal to number of members.");
      System.exit(1);
    }
    this.nbOfMembers = members;
    this.nbOfSenders = senders;
    this.nbOfMessages = messages;
    this.msgSize = size;
    try
    {
      initializeChannel();
    }
    catch (Exception e)
    {
      e.printStackTrace();
      System.out.println("Failed to initialize channel. Exiting ...");
      System.exit(1);
    }
  }

  /**
   * Initializes group communication channel and adapter.
   * 
   * @throws Exception if an error occurs
   */
  private void initializeChannel() throws Exception
  {
    final InetAddress MULTICAST_ADDRESS = InetAddress.getByName("224.7.65.23");
    final int MULTICAST_PORT = 2288;
    final IpAddress MULTICAST_IP = new IpAddress(MULTICAST_ADDRESS,
        MULTICAST_PORT);
    final InetAddress REPLY_ADDRESS = InetAddress.getLocalHost();
    final int REPLY_PORT = 0; // Let the system choose a port
    final IpAddress REPLY_IP = new IpAddress(REPLY_ADDRESS, REPLY_PORT);

    System.out.println("Using address: " + REPLY_IP);

    discovery = new UdpDiscoveryService(MULTICAST_IP, REPLY_IP);
    gms = new GroupMembershipService(REPLY_IP, TcpChannelPool.getChannelPool(),
        discovery);
    channel = new ReliableGroupChannelWithGms(gms);
    gms.registerGroupMembershipListener(this);

    // Join

    try
    {
      channel.join(testGid);
    }
    catch (AlreadyMemberException e)
    {
      e.printStackTrace();
      throw e;
    }

    adapter = new PullPushAdapter(channel, this);
  }

  /**
   * Send the messages to the group as defined by the command line parameters
   */
  private void sendMessages()
  {
    synchronized (this)
    {
      if (sending)
        return;
      sending = true;
    }
    ArrayList members = gms.getGroup(testGid).getMembers();
    int size = members.size();
    for (int i = 0; i < nbOfSenders; i++)
    {
      if (members.get(i).equals(channel.getLocalMembership()))
      { // Ok, we are a sender
        ByteMessage msg = new ByteMessage(new byte[msgSize]);
        System.out.println("Starting sending " + nbOfMessages + " message of "
            + msgSize + " bytes to " + size + " members.");
        System.out.println("Members are: "
            + gms.getGroup(testGid).getStringMembers());
        start = System.currentTimeMillis();
        for (int j = 0; j < nbOfMessages; j++)
        {
          try
          {
            channel.send(msg, members);
          }
          catch (Exception e)
          {
            e.printStackTrace();
            System.out.println("Failure while sending message " + j);
            System.exit(1);
          }
        }
        long end = System.currentTimeMillis();
        System.out.println("Average time per message: "
            + ((double) (end - start) / (double) (nbOfMessages)) + "ms");
        return;
      }
      else
        System.out.println("Member " + members.get(i) + " does not match "
            + channel.getLocalMembership());
    }
    System.out.println("I am not a sender, waiting for messages.");
  }

  //
  // Message listener interface
  //

  /**
   * @see org.objectweb.tribe.messages.MessageListener#receive(java.io.Serializable)
   */
  public void receive(Serializable msg)
  {
    if (msg instanceof PerfTestDone)
    {
      done++;
      if (done == nbOfMembers)
      {
        if (start != 0)
        {
          long end = System.currentTimeMillis();
          System.out.println("Overall average time per message: "
              + ((double) (end - start) / (double) (nbOfMessages)) + "ms");
        }
        System.out.println("Everybody is done, exiting ...");
        System.exit(0);
      }
      else
        System.out.println("Member " + ((PerfTestDone) msg).getSender()
            + " is done.");
      return;
    }

    if (msg instanceof ByteMessage)
    {
      received++;
      if (received == nbOfMessages * nbOfSenders)
      {
        try
        {
          channel.send(new PerfTestDone(channel.getLocalMembership().toString()));
        }
        catch (Exception e)
        {
          System.out.println("Failed to send last acknowledgement");
          e.printStackTrace();
        }
        System.out.println("Received " + (nbOfMessages * nbOfSenders)
            + " messages, acknowledging.");
      }
      else if (received % 1000 == 0)
      {
        System.out.println("Received " + received + " messages");
      }
    }
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#joinMember(org.objectweb.tribe.common.Member,
   *      org.objectweb.tribe.common.GroupIdentifier)
   */
  public void joinMember(Member m, GroupIdentifier gid)
  {
    System.out.println("Member " + m + " has joined the group " + gid);
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#quitMember(org.objectweb.tribe.common.Member,
   *      org.objectweb.tribe.common.GroupIdentifier)
   */
  public void quitMember(Member m, GroupIdentifier gid)
  {
    System.out.println("Member " + m + " has left the group " + gid);
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#groupComposition(org.objectweb.tribe.common.Group,
   *      org.objectweb.tribe.common.Address)
   */
  public void groupComposition(Group g, Address sender)
  {
    System.out.println(sender + " send group " + g.getGroupIdentifier()
        + " composition.");
    if (gms.getGroup(testGid).getMembers().size() == this.nbOfMembers)
      sendMessages();
    else
      System.out.println("Group now has "
          + gms.getGroup(testGid).getMembers().size() + " members.");
  }

  /**
   * @see org.objectweb.tribe.gms.GroupMembershipListener#failedMember(org.objectweb.tribe.common.Member,
   *      org.objectweb.tribe.common.GroupIdentifier,
   *      org.objectweb.tribe.common.Member)
   */
  public void failedMember(Member failed, GroupIdentifier gid, Member sender)
  {
    System.out.println("Member " + failed + " failed in group " + gid
        + "(reported by " + sender + ")");
  }

  //
  // Main method
  //

  /**
   * Main method taking the following arguments:
   * 
   * <pre>
   * 1. number of group members
   * 2. number of senders (first to join the group will send)
   * 3. number of messages to send to the group
   * 4. message msgSize in bytes
   * </pre>
   * 
   * @param args command line arguments
   */
  public static void main(String args[])
  {
    try
    {
      ChannelPerformaceTest perf = new ChannelPerformaceTest(Integer
          .parseInt(args[0]), Integer.parseInt(args[1]), Integer
          .parseInt(args[2]), Integer.parseInt(args[3]));
    }
    catch (Exception e)
    {
      e.printStackTrace();
      System.out
          .println("Usage: java org.objectweb.tribe.channel.ChannelPerformaceTest members senders messages msgSize");
      System.out.println("members: number of group members");
      System.out
          .println("senders: number of senders (first members to join the group will send)");
      System.out.println("messages: number of messages to send to the group");
      System.out.println("msgSize: message size in bytes\n");
      System.out
          .println("Example: java org.objectweb.tribe.channel.ChannelPerformaceTest 2 1 10000 1024");
      System.exit(1);
    }
  }

}