/**
 * Tribe: Group communication library.
 * Copyright (C) 2004 French National Institute For Research In Computer
 * Science And Control (INRIA).
 * Contact: tribe@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Initial developer(s): Bela Ban.
 * Contributor(s): Emmanuel Cecchet.
 */

package org.objectweb.tribe.demos.draw;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.net.InetAddress;
import java.util.Random;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;

import org.objectweb.tribe.channel.ReliableGroupChannelWithGms;
import org.objectweb.tribe.channel.tcp.TcpChannelPool;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.gms.GroupMembershipService;
import org.objectweb.tribe.gms.discovery.UdpDiscoveryService;

/**
 * Shared whiteboard, each new instance joins the same group. Each instance
 * chooses a random color, mouse moves are broadcast to all group members, which
 * then apply them to their canvas
 * <p>
 * 
 * @author Bela Ban, Oct 17 2001
 * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
 * @version 1.0
 */
public class WhiteBoard implements ActionListener
{
  private static final String         GROUP_NAME      = "tribe.whiteboard";
  private ReliableGroupChannelWithGms channel         = null;
  private int                         memberSize      = 1;
  boolean                             first           = true,
      cummulative = true;
  private JFrame                      mainFrame       = null;
  private JPanel                      subPanel        = null;
  private DrawPanel                   panel           = null;
  private JButton                     clearButton;
  private JButton                     leaveButton;
  private Random                      random          = new Random(System
                                                          .currentTimeMillis());
  private final Font                  defaultFont     = new Font("Helvetica",
                                                          Font.PLAIN, 12);
  private Color                       drawColor       = selectColor();
  private Color                       backgroundColor = Color.white;

  /**
   * Creates a new <code>WhiteBoard</code> object
   * 
   * @throws Exception if an error occurs
   */
  public WhiteBoard() throws Exception
  {
    final InetAddress MULTICAST_ADDRESS = InetAddress.getByName("224.7.65.23");
    final int MULTICAST_PORT = 2288;
    final IpAddress MULTICAST_IP = new IpAddress(MULTICAST_ADDRESS,
        MULTICAST_PORT);
    final InetAddress REPLY_ADDRESS = InetAddress.getLocalHost();
    final int REPLY_PORT = 0;
    final IpAddress REPLY_IP = new IpAddress(REPLY_ADDRESS, REPLY_PORT);

    UdpDiscoveryService discovery = new UdpDiscoveryService(MULTICAST_IP,
        REPLY_IP);
    GroupMembershipService gms = new GroupMembershipService(REPLY_IP,
        TcpChannelPool.getChannelPool(), discovery);

    channel = new ReliableGroupChannelWithGms(gms);
    //    channel.setChannelListener(this);
  }

  /**
   * Starts the WhiteBoard application. No parameter are required.
   * 
   * @param args not used
   */
  public static void main(String[] args)
  {
    WhiteBoard whiteBoard = null;

    try
    {
      whiteBoard = new WhiteBoard();
      whiteBoard.go();
    }
    catch (Throwable e)
    {
      e.printStackTrace();
      System.exit(0);
    }
  }

  /**
   * Randomly chooses a color.
   * 
   * @return a new Color
   */
  private Color selectColor()
  {
    int red = (Math.abs(random.nextInt()) % 255);
    int green = (Math.abs(random.nextInt()) % 255);
    int blue = (Math.abs(random.nextInt()) % 255);
    return new Color(red, green, blue);
  }

  /**
   * Join the group, create the frame and listen for events.
   * 
   * @throws Exception if an error occurs
   */
  public void go() throws Exception
  {
    channel.join(new GroupIdentifier(GROUP_NAME));

    mainFrame = new JFrame();
    mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    panel = new DrawPanel();
    panel.setBackground(backgroundColor);
    subPanel = new JPanel();
    mainFrame.getContentPane().add("Center", panel);
    clearButton = new JButton("Clear");
    clearButton.setFont(defaultFont);
    clearButton.addActionListener(this);
    leaveButton = new JButton("Leave & Exit");
    leaveButton.setFont(defaultFont);
    leaveButton.addActionListener(this);
    subPanel.add("South", clearButton);
    subPanel.add("South", leaveButton);
    mainFrame.getContentPane().add("South", subPanel);
    mainFrame.setBackground(backgroundColor);
    clearButton.setForeground(Color.blue);
    leaveButton.setForeground(Color.blue);
    setTitle();
    mainFrame.pack();
    mainFrame.setLocation(15, 25);
    mainFrame.setVisible(true);
    mainLoop();
  }

  /**
   * Set the windows title.
   * 
   * @param title title to set.
   */
  void setTitle(String title)
  {
    String tmp = "";
    if (title != null)
    {
      mainFrame.setTitle(title);
    }
    else
    {
      //      if (channel.getLocalAddress() != null)
      //        tmp += channel.getLocalAddress();
      tmp += " (" + memberSize + ") mbrs";
      mainFrame.setTitle(tmp);
    }
  }

  /**
   * Set the windows title
   */
  void setTitle()
  {
    setTitle(null);
  }

  /**
   * Main loop handling incoming messages.
   */
  public void mainLoop()
  {
    Object msg = null;
    WhiteBoardCommand comm;
    boolean fl = true;

    while (fl)
    {
      try
      {
        msg = channel.receive();

        //        if (tmp instanceof View)
        //        {
        //          View v = (View) tmp;
        //          System.out.println("** View=" + v);
        //          memberSize = v.size();
        //          if (mainFrame != null)
        //            setTitle();
        //          continue;
        //        }

        comm = null;

        if (msg instanceof WhiteBoardCommand)
          comm = (WhiteBoardCommand) msg;
        else
        {
          if (msg != null)
            System.out.println("*** Draw.run(): msg is " + msg.getClass());
          else
            System.out.println("*** Draw.run(): msg is null");
          continue;
        }

        switch (comm.mode)
        {
          case WhiteBoardCommand.DRAW :
            if (panel != null)
              panel.drawPoint(comm);
            break;
          case WhiteBoardCommand.CLEAR :
            clearPanel();
            continue;
          default :
            System.err
                .println("***** Draw.run(): received invalid draw command "
                    + comm.mode);
            break;
        }

      }
      catch (Exception e)
      {
        System.err.println(e);
        continue;
      }
    }
  }

  /* --------------- Callbacks --------------- */

  /**
   * Clear the panel
   */
  public void clearPanel()
  {
    if (panel != null)
      panel.clear();
  }

  /**
   * Send the clear panel order to other members.
   */
  public void sendClearPanelMsg()
  {
    WhiteBoardCommand comm = new WhiteBoardCommand();

    try
    {
      channel.send(comm);
    }
    catch (Exception ex)
    {
      System.err.println(ex);
    }
  }

  /**
   * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
   */
  public void actionPerformed(ActionEvent e)
  {
    String command = e.getActionCommand();
    if (command.equals("Clear"))
      sendClearPanelMsg();
    else if (command.equals("Leave & Exit"))
    {
      try
      {
        channel.close();
      }
      catch (Exception ex)
      {
        System.err.println(ex);
      }
      mainFrame.setVisible(false);
      mainFrame.dispose();
      System.exit(0);
    }
    else
      System.out.println("Unknown action");
  }

  /**
   * This class defines a DrawPanel used for handling mouse events and display.
   * 
   * @author Bela Ban
   * @author <a href="mailto:Emmanuel.Cecchet@inria.fr">Emmanuel Cecchet </a>
   * @version 1.0
   */
  private class DrawPanel extends JPanel implements MouseMotionListener
  {
    Dimension preferred_size = new Dimension(235, 170);
    Image     img            = null;                   // for drawing pixels
    Dimension d, imgsize;
    Graphics  gr             = null;

    /**
     * Creates a new <code>DrawPanel</code> object
     */
    public DrawPanel()
    {
      createOffscreenImage();
      addMouseMotionListener(this);
      addComponentListener(new ComponentAdapter()
      {
        public void componentResized(ComponentEvent e)
        {
          if (getWidth() <= 0 || getHeight() <= 0)
            return;
          createOffscreenImage();
        }
      });
    }

    /**
     * Create an off screen image.
     */
    void createOffscreenImage()
    {
      d = getSize();
      if (img == null || imgsize == null || imgsize.width != d.width
          || imgsize.height != d.height)
      {
        img = createImage(d.width, d.height);
        if (img != null)
          gr = img.getGraphics();
        imgsize = d;
      }
    }

    //
    // -------------- MouseMotionListener interface ---------------
    // 

    /**
     * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
     */
    public void mouseDragged(MouseEvent e)
    {
      int x = e.getX(), y = e.getY();
      WhiteBoardCommand comm = new WhiteBoardCommand(x, y, drawColor.getRed(),
          drawColor.getGreen(), drawColor.getBlue());

      try
      {
        channel.send(comm);
        Thread.yield(); // gives the repainter some breath
      }
      catch (Exception ex)
      {
        System.err.println(ex);
      }
    }

    /**
     * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
     */
    public void mouseMoved(MouseEvent e)
    {
    }

    //
    // -------------- End of MouseMotionListener interface --------------
    //

    /**
     * Adds pixel to queue and calls repaint() whenever we have MAX_ITEMS pixels
     * in the queue or when MAX_TIME msecs have elapsed (whichever comes first).
     * The advantage compared to just calling repaint() after adding a pixel to
     * the queue is that repaint() can most often draw multiple points at the
     * same time.
     */
    public void drawPoint(WhiteBoardCommand c)
    {
      if (c == null || gr == null)
        return;
      gr.setColor(new Color(c.r, c.g, c.b));
      gr.fillOval(c.x, c.y, 10, 10);
      repaint();
    }

    /**
     * Clear the frame content
     */
    public void clear()
    {
      if (gr == null)
        return;
      gr.clearRect(0, 0, getSize().width, getSize().height);
      repaint();
    }

    /**
     * @see java.awt.Component#getPreferredSize()
     */
    public Dimension getPreferredSize()
    {
      return preferred_size;
    }

    /**
     * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
     */
    public void paintComponent(Graphics g)
    {
      super.paintComponent(g);
      if (img != null)
      {
        g.drawImage(img, 0, 0, null);
      }
    }

  }

}

